#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2023 The TARTRL Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""""""

import torch
import numpy as np
from collections import defaultdict
from torch.utils.data.sampler import BatchSampler, SubsetRandomSampler

from tartrl.buffers.utils.util import get_shape_from_obs_space, get_shape_from_act_space
from tartrl.buffers.utils.obs_data import ObsData

def _flatten(T, N, x):
    return x.reshape(T * N, *x.shape[2:])


def _flatten_v3(T, N, agent_num, x):
    return x.reshape(T * N * agent_num, *x.shape[3:])


def _cast(x):
    return x.transpose(1, 2, 0, 3).reshape(-1, *x.shape[3:])


def _cast_v3(x):
    return x.transpose(1, 0, 2, 3).reshape(-1, *x.shape[2:])


def _shuffle_agent_grid(x, y):
    rows = np.indices((x, y))[0]
    # cols = np.stack([np.random.permutation(y) for _ in range(x)])
    cols = np.stack([np.arange(y) for _ in range(x)])
    return rows, cols


class ReplayData(object):
    def __init__(self, args, num_agents, obs_space, share_obs_space, act_space, data_client, episode_length=None):
        if episode_length is None:
            episode_length = args.episode_length
        self.episode_length = episode_length

        self.n_rollout_threads = args.n_rollout_threads
        # self.hidden_size = args.hidden_size
        self.hidden_size = args.rnn_hidden_size
        self.recurrent_N = args.recurrent_N
        self.gamma = args.gamma
        self.gae_lambda = args.gae_lambda
        self._use_gae = args.use_gae
        self._use_popart = args.use_popart
        self._use_valuenorm = args.use_valuenorm
        self._use_proper_time_limits = args.use_proper_time_limits

        self._mixed_obs = False  # for mixed observation

        obs_shape = get_shape_from_obs_space(obs_space)
        share_obs_shape = get_shape_from_obs_space(share_obs_space)

        # for mixed observation
        if 'Dict' in obs_shape.__class__.__name__:
            self._mixed_obs = True

            self.obs = {}
            self.share_obs = {}

            for key in obs_shape:
                self.obs[key] = np.zeros(
                    (self.episode_length + 1, self.n_rollout_threads, num_agents, *obs_shape[key].shape),
                    dtype=np.float32)
            for key in share_obs_shape:
                self.share_obs[key] = np.zeros(
                    (self.episode_length + 1, self.n_rollout_threads, num_agents, *share_obs_shape[key].shape),
                    dtype=np.float32)
            self.obs = ObsData(self.obs)
            self.share_obs = ObsData(self.share_obs)

        else:
            # deal with special attn format
            if type(obs_shape[-1]) == list:
                obs_shape = obs_shape[:1]

            if type(share_obs_shape[-1]) == list:
                share_obs_shape = share_obs_shape[:1]

            self.share_obs = np.zeros((self.episode_length + 1, self.n_rollout_threads, num_agents, *share_obs_shape),
                                      dtype=np.float32)
            self.obs = np.zeros((self.episode_length + 1, self.n_rollout_threads, num_agents, *obs_shape),
                                dtype=np.float32)

        self.rnn_states = np.zeros(
            (self.episode_length + 1, self.n_rollout_threads, num_agents, self.recurrent_N, self.hidden_size),
            dtype=np.float32)
        self.rnn_states_critic = np.zeros_like(self.rnn_states)

        self.value_preds = np.zeros(
            (self.episode_length + 1, self.n_rollout_threads, num_agents, 1), dtype=np.float32)
        self.returns = np.zeros_like(self.value_preds)

        if act_space.__class__.__name__ == 'Discrete':
            self.available_actions = np.ones((self.episode_length + 1, self.n_rollout_threads, num_agents, act_space.n),
                                             dtype=np.float32)
        else:
            self.available_actions = None

        act_shape = get_shape_from_act_space(act_space)

        self.actions = np.zeros(
            (self.episode_length, self.n_rollout_threads, num_agents, act_shape), dtype=np.float32)
        self.action_log_probs = np.zeros(
            (self.episode_length, self.n_rollout_threads, num_agents, act_shape), dtype=np.float32)

        self.rewards = np.zeros(
            (self.episode_length, self.n_rollout_threads, num_agents, 1), dtype=np.float32)

        self.masks = np.ones((self.episode_length + 1, self.n_rollout_threads, num_agents, 1), dtype=np.float32)
        self.bad_masks = np.ones_like(self.masks)
        self.active_masks = np.ones_like(self.masks)

        self.step = 0

    def get_batch_data(
            self,
            data_name:str,
            step: int,
    ):
        assert hasattr(self,data_name)
        data = getattr(self,data_name)

        if isinstance(data,ObsData):
            return data.step_batch(step)
        else:
            return np.concatenate(data[step])

    def dict_insert(self, data):
        if self._mixed_obs:
            for key in self.share_obs.keys():
                self.share_obs[key][self.step + 1] = data['share_obs'][key].copy()
            for key in self.obs.keys():
                self.obs[key][self.step + 1] = data['obs'][key].copy()
        else:
            self.share_obs[self.step + 1] = data['share_obs'].copy()
            self.obs[self.step + 1] = data['obs'].copy()

        if 'rnn_states' in data:
            self.rnn_states[self.step + 1] = data['rnn_states'].copy()
        if 'rnn_states_critic' in data:
            self.rnn_states_critic[self.step + 1] = data['rnn_states_critic'].copy()
        if 'actions' in data:
            self.actions[self.step] = data['actions'].copy()
        if "action_log_probs" in data:
            self.action_log_probs[self.step] = data['action_log_probs'].copy()

        if 'value_preds' in data:
            self.value_preds[self.step] = data['value_preds'].copy()
        if 'rewards' in data:
            self.rewards[self.step] = data['rewards'].copy()
        if 'masks' in data:
            self.masks[self.step + 1] = data['masks'].copy()

        if 'bad_masks' in data:
            self.bad_masks[self.step + 1] = data['bad_masks'].copy()
        if 'active_masks' in data:
            self.active_masks[self.step + 1] = data['active_masks'].copy()
        if 'available_actions' in data:
            self.available_actions[self.step + 1] = data['available_actions'].copy()

        self.step = (self.step + 1) % self.episode_length

    def insert(self, share_obs, obs, rnn_states, rnn_states_critic, actions, action_log_probs,
               value_preds, rewards, masks, bad_masks=None, active_masks=None, available_actions=None):

        if self._mixed_obs:
            for key in self.share_obs.keys():
                self.share_obs[key][self.step + 1] = share_obs[key].copy()
            for key in self.obs.keys():
                self.obs[key][self.step + 1] = obs[key].copy()
        else:
            self.share_obs[self.step + 1] = share_obs.copy()
            self.obs[self.step + 1] = obs.copy()

        self.rnn_states[self.step + 1] = rnn_states.copy()
        self.rnn_states_critic[self.step + 1] = rnn_states_critic.copy()
        self.actions[self.step] = actions.copy()
        # print(self.action_log_probs[self.step],action_log_probs,action_log_probs.shape)
        # import pdb;pdb.set_trace()
        self.action_log_probs[self.step] = action_log_probs.copy()
        self.value_preds[self.step] = value_preds.copy()
        self.rewards[self.step] = rewards.copy()
        self.masks[self.step + 1] = masks.copy()
        if bad_masks is not None:
            self.bad_masks[self.step + 1] = bad_masks.copy()
        if active_masks is not None:
            self.active_masks[self.step + 1] = active_masks.copy()
        if available_actions is not None:
            self.available_actions[self.step + 1] = available_actions.copy()
        self.step = (self.step + 1) % self.episode_length

    def init_buffer(self, share_obs, obs, available_actions=None):
        if self._mixed_obs:
            for key in self.share_obs.keys():
                self.share_obs[key][0] = share_obs[key].copy()
            for key in self.obs.keys():
                self.obs[key][0] = obs[key].copy()
        else:
            self.share_obs[0] = share_obs.copy()
            self.obs[0] = obs.copy()
        if available_actions is not None and self.available_actions is not None:
            self.available_actions[0] = available_actions

    def chooseinsert(self, share_obs, obs, rnn_states, rnn_states_critic, actions, action_log_probs,
                     value_preds, rewards, masks, bad_masks=None, active_masks=None, available_actions=None):
        self.share_obs[self.step] = share_obs.copy()
        self.obs[self.step] = obs.copy()
        self.rnn_states[self.step + 1] = rnn_states.copy()
        self.rnn_states_critic[self.step + 1] = rnn_states_critic.copy()
        self.actions[self.step] = actions.copy()
        self.action_log_probs[self.step] = action_log_probs.copy()
        self.value_preds[self.step] = value_preds.copy()
        self.rewards[self.step] = rewards.copy()
        self.masks[self.step + 1] = masks.copy()
        if bad_masks is not None:
            self.bad_masks[self.step + 1] = bad_masks.copy()
        if active_masks is not None:
            self.active_masks[self.step] = active_masks.copy()
        if available_actions is not None:
            self.available_actions[self.step] = available_actions.copy()

        self.step = (self.step + 1) % self.episode_length

    def after_update(self):
        assert self.step == 0, "step:{} episode:{}".format(self.step, self.episode_length)
        if self._mixed_obs:
            for key in self.share_obs.keys():
                self.share_obs[key][0] = self.share_obs[key][-1].copy()
            for key in self.obs.keys():
                self.obs[key][0] = self.obs[key][-1].copy()
        else:
            self.share_obs[0] = self.share_obs[-1].copy()
            self.obs[0] = self.obs[-1].copy()
        self.rnn_states[0] = self.rnn_states[-1].copy()
        self.rnn_states_critic[0] = self.rnn_states_critic[-1].copy()
        self.masks[0] = self.masks[-1].copy()
        self.bad_masks[0] = self.bad_masks[-1].copy()
        self.active_masks[0] = self.active_masks[-1].copy()
        if self.available_actions is not None:
            self.available_actions[0] = self.available_actions[-1].copy()

    def chooseafter_update(self):
        self.rnn_states[0] = self.rnn_states[-1].copy()
        self.rnn_states_critic[0] = self.rnn_states_critic[-1].copy()
        self.masks[0] = self.masks[-1].copy()
        self.bad_masks[0] = self.bad_masks[-1].copy()

    def compute_returns(self, next_value, value_normalizer=None):
        if self._use_proper_time_limits:
            if self._use_gae:
                self.value_preds[-1] = next_value
                gae = 0
                for step in reversed(range(self.rewards.shape[0])):
                    if self._use_popart or self._use_valuenorm:
                        # step + 1
                        delta = self.rewards[step] + self.gamma * value_normalizer.denormalize(
                            self.value_preds[step + 1]) * self.masks[step + 1] \
                                - value_normalizer.denormalize(self.value_preds[step])
                        gae = delta + self.gamma * self.gae_lambda * gae * self.masks[step + 1]
                        gae = gae * self.bad_masks[step + 1]
                        self.returns[step] = gae + value_normalizer.denormalize(self.value_preds[step])
                    else:
                        delta = self.rewards[step] + self.gamma * self.value_preds[step + 1] * self.masks[step + 1] - \
                                self.value_preds[step]
                        gae = delta + self.gamma * self.gae_lambda * self.masks[step + 1] * gae
                        gae = gae * self.bad_masks[step + 1]
                        self.returns[step] = gae + self.value_preds[step]
            else:
                self.returns[-1] = next_value
                for step in reversed(range(self.rewards.shape[0])):
                    if self._use_popart or self._use_valuenorm:
                        self.returns[step] = (self.returns[step + 1] * self.gamma * self.masks[step + 1] + self.rewards[
                            step]) * self.bad_masks[step + 1] \
                                             + (1 - self.bad_masks[step + 1]) * value_normalizer.denormalize(
                            self.value_preds[step])
                    else:
                        self.returns[step] = (self.returns[step + 1] * self.gamma * self.masks[step + 1] + self.rewards[
                            step]) * self.bad_masks[step + 1] \
                                             + (1 - self.bad_masks[step + 1]) * self.value_preds[step]
        else:
            if self._use_gae:
                self.value_preds[-1] = next_value
                gae = 0
                for step in reversed(range(self.rewards.shape[0])):
                    if self._use_popart or self._use_valuenorm:
                        delta = self.rewards[step] + self.gamma * value_normalizer.denormalize(
                            self.value_preds[step + 1]) * self.masks[step + 1] \
                                - value_normalizer.denormalize(self.value_preds[step])
                        gae = delta + self.gamma * self.gae_lambda * self.masks[step + 1] * gae
                        self.returns[step] = gae + value_normalizer.denormalize(self.value_preds[step])
                    else:
                        delta = self.rewards[step] + self.gamma * self.value_preds[step + 1] * self.masks[step + 1] - \
                                self.value_preds[step]
                        gae = delta + self.gamma * self.gae_lambda * self.masks[step + 1] * gae
                        self.returns[step] = gae + self.value_preds[step]
            else:
                self.returns[-1] = next_value
                for step in reversed(range(self.rewards.shape[0])):
                    self.returns[step] = self.returns[step + 1] * self.gamma * self.masks[step + 1] + self.rewards[step]

    def feed_forward_generator(self, advantages, num_mini_batch=None, mini_batch_size=None,
                               share_obs_process_func=None):
        episode_length, n_rollout_threads, num_agents = self.rewards.shape[0:3]
        batch_size = n_rollout_threads * episode_length * num_agents

        if mini_batch_size is None:
            assert batch_size >= num_mini_batch, (
                "PPO requires the number of processes ({}) "
                "* number of steps ({}) * number of agents ({}) = {} "
                "to be greater than or equal to the number of PPO mini batches ({})."
                "".format(n_rollout_threads, episode_length, num_agents,
                          n_rollout_threads * episode_length * num_agents,
                          num_mini_batch))
            mini_batch_size = batch_size // num_mini_batch

        # rand = torch.randperm(batch_size).numpy()
        # sampler = [rand[i*mini_batch_size:(i+1)*mini_batch_size] for i in range(num_mini_batch)]
        sampler = BatchSampler(
            SubsetRandomSampler(range(batch_size)),
            mini_batch_size,
            drop_last=True)

        if self._mixed_obs:
            share_obs = {}
            obs = {}
            for key in self.share_obs.keys():
                share_obs[key] = self.share_obs[key][:-1].reshape(-1, *self.share_obs[key].shape[3:])
            for key in self.obs.keys():
                obs[key] = self.obs[key][:-1].reshape(-1, *self.obs[key].shape[3:])
        else:
            share_obs = self.share_obs[:-1].reshape(-1, *self.share_obs.shape[3:])
            obs = self.obs[:-1].reshape(-1, *self.obs.shape[3:])
        rnn_states = self.rnn_states[:-1].reshape(-1, *self.rnn_states.shape[3:])
        rnn_states_critic = self.rnn_states_critic[:-1].reshape(-1, *self.rnn_states_critic.shape[3:])
        actions = self.actions.reshape(-1, self.actions.shape[-1])
        if self.available_actions is not None:
            available_actions = self.available_actions[:-1].reshape(-1, self.available_actions.shape[-1])
        value_preds = self.value_preds[:-1].reshape(-1, 1)
        returns = self.returns[:-1].reshape(-1, 1)
        masks = self.masks[:-1].reshape(-1, 1)
        active_masks = self.active_masks[:-1].reshape(-1, 1)
        action_log_probs = self.action_log_probs.reshape(-1, self.action_log_probs.shape[-1])
        if advantages is not None:
            advantages = advantages.reshape(-1, 1)

        for indices in sampler:
            # obs size [T+1 N M Dim]-->[T N M Dim]-->[T*N*M,Dim]-->[index,Dim]
            if self._mixed_obs:
                share_obs_batch = {}
                obs_batch = {}
                for key in share_obs.keys():
                    share_obs_batch[key] = share_obs[key][indices]
                for key in obs.keys():
                    obs_batch[key] = obs[key][indices]
            else:
                share_obs_batch = share_obs[indices]
                obs_batch = obs[indices]
            rnn_states_batch = rnn_states[indices]
            rnn_states_critic_batch = rnn_states_critic[indices]
            actions_batch = actions[indices]
            if self.available_actions is not None:
                available_actions_batch = available_actions[indices]
            else:
                available_actions_batch = None
            value_preds_batch = value_preds[indices]
            return_batch = returns[indices]
            masks_batch = masks[indices]
            active_masks_batch = active_masks[indices]
            old_action_log_probs_batch = action_log_probs[indices]
            if advantages is None:
                adv_targ = None
            else:
                adv_targ = advantages[indices]
            if share_obs_process_func is not None:
                share_obs_batch = share_obs_process_func(share_obs_batch)

            yield share_obs_batch, obs_batch, rnn_states_batch, rnn_states_critic_batch, actions_batch, value_preds_batch, return_batch, masks_batch, active_masks_batch, old_action_log_probs_batch, adv_targ, available_actions_batch

    def feed_forward_share_obs_generator(self, advantages, num_mini_batch=None, mini_batch_size=None,
                                         share_obs_process_func=None):
        episode_length, n_rollout_threads, num_agents = self.rewards.shape[0:3]
        batch_size = n_rollout_threads * episode_length

        if mini_batch_size is None:
            assert batch_size >= num_mini_batch, (
                "PPO requires the number of processes ({}) "
                "* number of steps ({}) * number of agents ({}) = {} "
                "to be greater than or equal to the number of PPO mini batches ({})."
                "".format(n_rollout_threads, episode_length, num_agents, n_rollout_threads * episode_length,
                          num_mini_batch))
            mini_batch_size = batch_size // num_mini_batch

        # rand = torch.randperm(batch_size).numpy()
        # sampler = [rand[i*mini_batch_size:(i+1)*mini_batch_size] for i in range(num_mini_batch)]
        sampler = BatchSampler(
            SubsetRandomSampler(range(batch_size)),
            mini_batch_size,
            drop_last=True)

        if self._mixed_obs:
            share_obs = {}
            for key in self.share_obs.keys():
                share_obs[key] = self.share_obs[key][:-1].reshape(-1, *self.share_obs[key].shape[3:])
        else:
            share_obs = self.share_obs[:-1, :, 0].reshape(-1, *self.share_obs.shape[3:])  # [T*N,Dim]

        actions = self.actions[:, :, 0].reshape(-1, self.actions.shape[-1])

        for indices in sampler:
            # T is episode length, N is rollout number, M is agent number, dim is dimension
            # share_obs size [T+1 N M Dim]-->[T N Dim]-->[T*N,Dim]-->[index,Dim]
            if self._mixed_obs:
                share_obs_batch = {}
                for key in share_obs.keys():
                    share_obs_batch[key] = share_obs[key][indices]
            else:
                share_obs_batch = share_obs[indices]

            actions_batch = actions[indices]

            if share_obs_process_func is not None:
                share_obs_batch = share_obs_process_func(share_obs_batch)

            yield share_obs_batch, None, None, None, actions_batch, None, None, None, None, None, None, None

    def feed_forward_generator_transformer(self, advantages, num_mini_batch=None, mini_batch_size=None):
        """
        Yield training data for MLP policies.
        :param advantages: (np.ndarray) advantage estimates.
        :param num_mini_batch: (int) number of minibatches to split the batch into.
        :param mini_batch_size: (int) number of samples in each minibatch.
        """
        episode_length, n_rollout_threads, num_agents = self.rewards.shape[0:3]
        batch_size = n_rollout_threads * episode_length

        if mini_batch_size is None:
            assert batch_size >= num_mini_batch, (
                "PPO requires the number of processes ({}) "
                "* number of steps ({}) = {} "
                "to be greater than or equal to the number of PPO mini batches ({})."
                "".format(n_rollout_threads, episode_length,
                          n_rollout_threads * episode_length,
                          num_mini_batch))
            mini_batch_size = batch_size // num_mini_batch

        rand = torch.randperm(batch_size).numpy()
        sampler = [rand[i * mini_batch_size:(i + 1) * mini_batch_size] for i in range(num_mini_batch)]
        rows, cols = _shuffle_agent_grid(batch_size, num_agents)

        # keep (num_agent, dim)
        share_obs = self.share_obs[:-1].reshape(-1, *self.share_obs.shape[2:])
        share_obs = share_obs[rows, cols]
        obs = self.obs[:-1].reshape(-1, *self.obs.shape[2:])
        obs = obs[rows, cols]
        rnn_states = self.rnn_states[:-1].reshape(-1, *self.rnn_states.shape[2:])
        rnn_states = rnn_states[rows, cols]
        rnn_states_critic = self.rnn_states_critic[:-1].reshape(-1, *self.rnn_states_critic.shape[2:])
        rnn_states_critic = rnn_states_critic[rows, cols]
        actions = self.actions.reshape(-1, *self.actions.shape[2:])
        import pdb;
        pdb.set_trace()

        actions = actions[rows, cols]

        if self.available_actions is not None:
            available_actions = self.available_actions[:-1].reshape(-1, *self.available_actions.shape[2:])
            available_actions = available_actions[rows, cols]
        value_preds = self.value_preds[:-1].reshape(-1, *self.value_preds.shape[2:])
        value_preds = value_preds[rows, cols]
        returns = self.returns[:-1].reshape(-1, *self.returns.shape[2:])
        returns = returns[rows, cols]
        masks = self.masks[:-1].reshape(-1, *self.masks.shape[2:])
        masks = masks[rows, cols]
        active_masks = self.active_masks[:-1].reshape(-1, *self.active_masks.shape[2:])
        active_masks = active_masks[rows, cols]
        action_log_probs = self.action_log_probs.reshape(-1, *self.action_log_probs.shape[2:])
        action_log_probs = action_log_probs[rows, cols]
        advantages = advantages.reshape(-1, *advantages.shape[2:])
        advantages = advantages[rows, cols]

        for indices in sampler:
            # [L,T,N,Dim]-->[L*T,N,Dim]-->[index,N,Dim]-->[index*N, Dim]
            share_obs_batch = share_obs[indices].reshape(-1, *share_obs.shape[2:])
            obs_batch = obs[indices].reshape(-1, *obs.shape[2:])
            rnn_states_batch = rnn_states[indices].reshape(-1, *rnn_states.shape[2:])
            rnn_states_critic_batch = rnn_states_critic[indices].reshape(-1, *rnn_states_critic.shape[2:])
            actions_batch = actions[indices].reshape(-1, *actions.shape[2:])
            if self.available_actions is not None:
                available_actions_batch = available_actions[indices].reshape(-1, *available_actions.shape[2:])
            else:
                available_actions_batch = None
            value_preds_batch = value_preds[indices].reshape(-1, *value_preds.shape[2:])
            return_batch = returns[indices].reshape(-1, *returns.shape[2:])
            masks_batch = masks[indices].reshape(-1, *masks.shape[2:])
            active_masks_batch = active_masks[indices].reshape(-1, *active_masks.shape[2:])
            old_action_log_probs_batch = action_log_probs[indices].reshape(-1, *action_log_probs.shape[2:])
            if advantages is None:
                adv_targ = None
            else:
                adv_targ = advantages[indices].reshape(-1, *advantages.shape[2:])

            yield share_obs_batch, obs_batch, rnn_states_batch, rnn_states_critic_batch, actions_batch, \
                value_preds_batch, return_batch, masks_batch, active_masks_batch, old_action_log_probs_batch, \
                adv_targ, available_actions_batch

    def naive_recurrent_generator(self, advantages, num_mini_batch):
        episode_length, n_rollout_threads, num_agents = self.rewards.shape[0:3]
        batch_size = n_rollout_threads * num_agents
        assert n_rollout_threads * num_agents >= num_mini_batch, (
            "PPO requires the number of processes ({})* number of agents ({}) "
            "to be greater than or equal to the number of "
            "PPO mini batches ({}).".format(n_rollout_threads, num_agents, num_mini_batch))
        num_envs_per_batch = batch_size // num_mini_batch
        perm = torch.randperm(batch_size).numpy()

        if self._mixed_obs:
            share_obs = {}
            obs = {}
            for key in self.share_obs.keys():
                share_obs[key] = self.share_obs[key].reshape(-1, batch_size, *self.share_obs[key].shape[3:])
            for key in self.obs.keys():
                obs[key] = self.obs[key].reshape(-1, batch_size, *self.obs[key].shape[3:])
        else:
            share_obs = self.share_obs.reshape(-1, batch_size, *self.share_obs.shape[3:])
            obs = self.obs.reshape(-1, batch_size, *self.obs.shape[3:])
        rnn_states = self.rnn_states.reshape(-1, batch_size, *self.rnn_states.shape[3:])
        rnn_states_critic = self.rnn_states_critic.reshape(-1, batch_size, *self.rnn_states_critic.shape[3:])
        actions = self.actions.reshape(-1, batch_size, self.actions.shape[-1])
        if self.available_actions is not None:
            available_actions = self.available_actions.reshape(-1, batch_size, self.available_actions.shape[-1])
        value_preds = self.value_preds.reshape(-1, batch_size, 1)
        returns = self.returns.reshape(-1, batch_size, 1)
        masks = self.masks.reshape(-1, batch_size, 1)
        active_masks = self.active_masks.reshape(-1, batch_size, 1)
        action_log_probs = self.action_log_probs.reshape(-1, batch_size, self.action_log_probs.shape[-1])
        advantages = advantages.reshape(-1, batch_size, 1)

        for start_ind in range(0, batch_size, num_envs_per_batch):

            if self._mixed_obs:
                share_obs_batch = defaultdict(list)
                obs_batch = defaultdict(list)
            else:
                share_obs_batch = []
                obs_batch = []

            rnn_states_batch = []
            rnn_states_critic_batch = []
            actions_batch = []
            available_actions_batch = []
            value_preds_batch = []
            return_batch = []
            masks_batch = []
            active_masks_batch = []
            old_action_log_probs_batch = []
            adv_targ = []

            for offset in range(num_envs_per_batch):
                ind = perm[start_ind + offset]
                if self._mixed_obs:
                    for key in share_obs.keys():
                        share_obs_batch[key].append(share_obs[key][:-1, ind])
                    for key in obs.keys():
                        obs_batch[key].append(obs[key][:-1, ind])
                else:
                    share_obs_batch.append(share_obs[:-1, ind])
                    obs_batch.append(obs[:-1, ind])
                rnn_states_batch.append(rnn_states[0:1, ind])
                rnn_states_critic_batch.append(rnn_states_critic[0:1, ind])
                actions_batch.append(actions[:, ind])
                if self.available_actions is not None:
                    available_actions_batch.append(available_actions[:-1, ind])
                value_preds_batch.append(value_preds[:-1, ind])
                return_batch.append(returns[:-1, ind])
                masks_batch.append(masks[:-1, ind])
                active_masks_batch.append(active_masks[:-1, ind])
                old_action_log_probs_batch.append(action_log_probs[:, ind])
                adv_targ.append(advantages[:, ind])

            # [N[T, dim]]
            T, N = self.episode_length, num_envs_per_batch
            # These are all from_numpys of size (T, N, -1)
            if self._mixed_obs:
                for key in share_obs_batch.keys():
                    share_obs_batch[key] = np.stack(share_obs_batch[key], 1)
                for key in obs_batch.keys():
                    obs_batch[key] = np.stack(obs_batch[key], 1)
            else:
                share_obs_batch = np.stack(share_obs_batch, 1)
                obs_batch = np.stack(obs_batch, 1)
            actions_batch = np.stack(actions_batch, 1)
            if self.available_actions is not None:
                available_actions_batch = np.stack(available_actions_batch, 1)
            value_preds_batch = np.stack(value_preds_batch, 1)
            return_batch = np.stack(return_batch, 1)
            masks_batch = np.stack(masks_batch, 1)
            active_masks_batch = np.stack(active_masks_batch, 1)
            old_action_log_probs_batch = np.stack(old_action_log_probs_batch, 1)
            adv_targ = np.stack(adv_targ, 1)

            # States is just a (N, dim) from_numpy [N[1,dim]]
            rnn_states_batch = np.stack(rnn_states_batch).reshape(N, *self.rnn_states.shape[3:])
            rnn_states_critic_batch = np.stack(rnn_states_critic_batch).reshape(N, *self.rnn_states_critic.shape[3:])

            # Flatten the (T, N, ...) from_numpys to (T * N, ...)
            if self._mixed_obs:
                for key in share_obs_batch.keys():
                    share_obs_batch[key] = _flatten(T, N, share_obs_batch[key])
                for key in obs_batch.keys():
                    obs_batch[key] = _flatten(T, N, obs_batch[key])
            else:
                share_obs_batch = _flatten(T, N, share_obs_batch)
                obs_batch = _flatten(T, N, obs_batch)
            actions_batch = _flatten(T, N, actions_batch)
            if self.available_actions is not None:
                available_actions_batch = _flatten(T, N, available_actions_batch)
            else:
                available_actions_batch = None
            value_preds_batch = _flatten(T, N, value_preds_batch)
            return_batch = _flatten(T, N, return_batch)
            masks_batch = _flatten(T, N, masks_batch)
            active_masks_batch = _flatten(T, N, active_masks_batch)
            old_action_log_probs_batch = _flatten(T, N, old_action_log_probs_batch)
            adv_targ = _flatten(T, N, adv_targ)

            yield share_obs_batch, obs_batch, rnn_states_batch, rnn_states_critic_batch, actions_batch, value_preds_batch, return_batch, masks_batch, active_masks_batch, old_action_log_probs_batch, adv_targ, available_actions_batch

    def recurrent_generator_v2(self, advantages, num_mini_batch=None, mini_batch_size=None):
        """
        Yield training data for MLP policies.
        :param advantages: (np.ndarray) advantage estimates.
        :param num_mini_batch: (int) number of minibatches to split the batch into.
        :param mini_batch_size: (int) number of samples in each minibatch.
        """
        episode_length, n_rollout_threads, num_agents = self.rewards.shape[0:3]
        batch_size = n_rollout_threads * episode_length

        if mini_batch_size is None:
            assert batch_size >= num_mini_batch, (
                "PPO requires the number of processes ({}) "
                "* number of steps ({}) = {} "
                "to be greater than or equal to the number of PPO mini batches ({})."
                "".format(n_rollout_threads, episode_length,
                          n_rollout_threads * episode_length,
                          num_mini_batch))
            mini_batch_size = batch_size // num_mini_batch

        rand = torch.randperm(batch_size).numpy()
        sampler = [rand[i * mini_batch_size:(i + 1) * mini_batch_size] for i in range(num_mini_batch)]

        # keep (num_agent, dim)
        share_obs = self.share_obs[:-1].reshape(-1, *self.share_obs.shape[2:])

        obs = self.obs[:-1].reshape(-1, *self.obs.shape[2:])

        rnn_states = self.rnn_states[:-1].reshape(-1, *self.rnn_states.shape[2:])

        rnn_states_critic = self.rnn_states_critic[:-1].reshape(-1, *self.rnn_states_critic.shape[2:])

        actions = self.actions.reshape(-1, *self.actions.shape[2:])

        if self.available_actions is not None:
            available_actions = self.available_actions[:-1].reshape(-1, *self.available_actions.shape[2:])

        value_preds = self.value_preds[:-1].reshape(-1, *self.value_preds.shape[2:])

        returns = self.returns[:-1].reshape(-1, *self.returns.shape[2:])

        masks = self.masks[:-1].reshape(-1, *self.masks.shape[2:])

        active_masks = self.active_masks[:-1].reshape(-1, *self.active_masks.shape[2:])

        action_log_probs = self.action_log_probs.reshape(-1, *self.action_log_probs.shape[2:])

        advantages = advantages.reshape(-1, *advantages.shape[2:])

        shuffle = False
        if shuffle:
            rows, cols = _shuffle_agent_grid(batch_size, num_agents)

            if self.available_actions is not None:
                available_actions = available_actions[rows, cols]
            share_obs = share_obs[rows, cols]
            obs = obs[rows, cols]
            rnn_states = rnn_states[rows, cols]
            rnn_states_critic = rnn_states_critic[rows, cols]
            actions = actions[rows, cols]
            value_preds = value_preds[rows, cols]
            returns = returns[rows, cols]
            masks = masks[rows, cols]
            active_masks = active_masks[rows, cols]
            action_log_probs = action_log_probs[rows, cols]
            advantages = advantages[rows, cols]

        for indices in sampler:
            # [L,T,N,Dim]-->[L*T,N,Dim]-->[index,N,Dim]-->[index*N, Dim]
            share_obs_batch = share_obs[indices].reshape(-1, *share_obs.shape[2:])
            obs_batch = obs[indices].reshape(-1, *obs.shape[2:])
            rnn_states_batch = rnn_states[indices].reshape(-1, *rnn_states.shape[2:])
            rnn_states_critic_batch = rnn_states_critic[indices].reshape(-1, *rnn_states_critic.shape[2:])
            actions_batch = actions[indices].reshape(-1, *actions.shape[2:])
            if self.available_actions is not None:
                available_actions_batch = available_actions[indices].reshape(-1, *available_actions.shape[2:])
            else:
                available_actions_batch = None
            value_preds_batch = value_preds[indices].reshape(-1, *value_preds.shape[2:])
            return_batch = returns[indices].reshape(-1, *returns.shape[2:])
            masks_batch = masks[indices].reshape(-1, *masks.shape[2:])
            active_masks_batch = active_masks[indices].reshape(-1, *active_masks.shape[2:])
            old_action_log_probs_batch = action_log_probs[indices].reshape(-1, *action_log_probs.shape[2:])
            if advantages is None:
                adv_targ = None
            else:
                adv_targ = advantages[indices].reshape(-1, *advantages.shape[2:])
            yield share_obs_batch, obs_batch, rnn_states_batch, rnn_states_critic_batch, actions_batch, \
                value_preds_batch, return_batch, masks_batch, active_masks_batch, old_action_log_probs_batch, \
                adv_targ, available_actions_batch

    def recurrent_generator_v3(self, advantages, num_mini_batch, data_chunk_length):
        episode_length, n_rollout_threads, num_agents = self.rewards.shape[0:3]
        batch_size = n_rollout_threads * episode_length
        data_chunks = batch_size // data_chunk_length  # [C=r*T*M/L]
        mini_batch_size = data_chunks // num_mini_batch

        assert n_rollout_threads * episode_length >= data_chunk_length, (
            "PPO requires the nfumber of processes ({}) * episode length ({}) "
            "to be greater than or equal to the number of "
            "data chunk length ({}).".format(n_rollout_threads, num_agents, episode_length, data_chunk_length))

        rand = torch.randperm(data_chunks).numpy()
        sampler = [rand[i * mini_batch_size:(i + 1) * mini_batch_size] for i in range(num_mini_batch)]

        share_obs = _cast_v3(self.share_obs[:-1])
        obs = _cast_v3(self.obs[:-1])

        actions = _cast_v3(self.actions)
        action_log_probs = _cast_v3(self.action_log_probs)
        advantages = _cast_v3(advantages)
        value_preds = _cast_v3(self.value_preds[:-1])
        returns = _cast_v3(self.returns[:-1])
        masks = _cast_v3(self.masks[:-1])
        active_masks = _cast_v3(self.active_masks[:-1])

        rnn_states = self.rnn_states[:-1].transpose(1, 0, 2, 3, 4).reshape(-1, *self.rnn_states.shape[2:])
        rnn_states_critic = self.rnn_states_critic[:-1].transpose(1, 0, 2, 3, 4).reshape(-1,
                                                                                         *self.rnn_states_critic.shape[
                                                                                          2:])

        if self.available_actions is not None:
            available_actions = _cast_v3(self.available_actions[:-1])

        for indices in sampler:
            share_obs_batch = []
            obs_batch = []

            rnn_states_batch = []
            rnn_states_critic_batch = []
            actions_batch = []
            available_actions_batch = []
            value_preds_batch = []
            return_batch = []
            masks_batch = []
            active_masks_batch = []
            old_action_log_probs_batch = []
            adv_targ = []

            for index in indices:

                ind = index * data_chunk_length
                # size [T+1 N M Dim]-->[T N M Dim]-->[N,M,T,Dim]-->[N*M*T,Dim]-->[L,Dim]
                # [L, agent_num, Dim]
                share_obs_batch.append(share_obs[ind:ind + data_chunk_length])
                obs_batch.append(obs[ind:ind + data_chunk_length])

                actions_batch.append(actions[ind:ind + data_chunk_length])
                if self.available_actions is not None:
                    available_actions_batch.append(available_actions[ind:ind + data_chunk_length])
                value_preds_batch.append(value_preds[ind:ind + data_chunk_length])
                return_batch.append(returns[ind:ind + data_chunk_length])
                masks_batch.append(masks[ind:ind + data_chunk_length])
                active_masks_batch.append(active_masks[ind:ind + data_chunk_length])
                old_action_log_probs_batch.append(action_log_probs[ind:ind + data_chunk_length])
                adv_targ.append(advantages[ind:ind + data_chunk_length])
                # size [T+1 N M Dim]-->[T N M Dim]-->[N M T Dim]-->[N*M*T,Dim]-->[1,Dim]
                # [1,agent_num, Dim]
                rnn_states_batch.append(rnn_states[ind])
                rnn_states_critic_batch.append(rnn_states_critic[ind])

            L, N = data_chunk_length, mini_batch_size

            # These are all from_numpys of size (L, N, agent_num, Dim)

            share_obs_batch = np.stack(share_obs_batch, axis=1)
            obs_batch = np.stack(obs_batch, axis=1)

            actions_batch = np.stack(actions_batch, axis=1)
            if self.available_actions is not None:
                available_actions_batch = np.stack(available_actions_batch, axis=1)
            value_preds_batch = np.stack(value_preds_batch, axis=1)
            return_batch = np.stack(return_batch, axis=1)
            masks_batch = np.stack(masks_batch, axis=1)
            active_masks_batch = np.stack(active_masks_batch, axis=1)
            old_action_log_probs_batch = np.stack(old_action_log_probs_batch, axis=1)
            adv_targ = np.stack(adv_targ, axis=1)

            # States is just a (N, agent_num, -1) from_numpy
            rnn_states_batch = np.stack(rnn_states_batch).reshape(N * num_agents, *self.rnn_states.shape[3:])
            rnn_states_critic_batch = np.stack(rnn_states_critic_batch).reshape(N * num_agents,
                                                                                *self.rnn_states_critic.shape[3:])
            # import pdb;pdb.set_trace()
            # Flatten the (L, N, ...) from_numpys to (L * N, ...)

            share_obs_batch = _flatten_v3(L, N, num_agents, share_obs_batch)
            obs_batch = _flatten_v3(L, N, num_agents, obs_batch)
            actions_batch = _flatten_v3(L, N, num_agents, actions_batch)
            if self.available_actions is not None:
                available_actions_batch = _flatten_v3(L, N, num_agents, available_actions_batch)
            else:
                available_actions_batch = None
            value_preds_batch = _flatten_v3(L, N, num_agents, value_preds_batch)
            return_batch = _flatten_v3(L, N, num_agents, return_batch)
            masks_batch = _flatten_v3(L, N, num_agents, masks_batch)
            active_masks_batch = _flatten_v3(L, N, num_agents, active_masks_batch)
            old_action_log_probs_batch = _flatten_v3(L, N, num_agents, old_action_log_probs_batch)
            adv_targ = _flatten_v3(L, N, num_agents, adv_targ)
            # print(L,N,num_agents)
            # import pdb;pdb.set_trace()
            yield share_obs_batch, obs_batch, rnn_states_batch, rnn_states_critic_batch, actions_batch, value_preds_batch, return_batch, masks_batch, active_masks_batch, old_action_log_probs_batch, adv_targ, available_actions_batch

    def recurrent_generator(self, advantages, num_mini_batch, data_chunk_length):
        episode_length, n_rollout_threads, num_agents = self.rewards.shape[0:3]
        batch_size = n_rollout_threads * episode_length * num_agents
        data_chunks = batch_size // data_chunk_length  # [C=r*T*M/L]
        mini_batch_size = data_chunks // num_mini_batch

        assert n_rollout_threads * episode_length * num_agents >= data_chunk_length, (
            "PPO requires the number of processes ({})* number of agents ({}) * episode length ({}) "
            "to be greater than or equal to the number of "
            "data chunk length ({}).".format(n_rollout_threads, num_agents, episode_length, data_chunk_length))

        rand = torch.randperm(data_chunks).numpy()
        sampler = [rand[i * mini_batch_size:(i + 1) * mini_batch_size] for i in range(num_mini_batch)]

        if self._mixed_obs:
            share_obs = {}
            obs = {}
            for key in self.share_obs.keys():
                if len(self.share_obs[key].shape) == 6:
                    share_obs[key] = self.share_obs[key][:-1].transpose(1, 2, 0, 3, 4, 5).reshape(-1, *self.share_obs[
                                                                                                           key].shape[
                                                                                                       3:])
                elif len(self.share_obs[key].shape) == 5:
                    share_obs[key] = self.share_obs[key][:-1].transpose(1, 2, 0, 3, 4).reshape(-1, *self.share_obs[
                                                                                                        key].shape[3:])
                else:
                    share_obs[key] = _cast(self.share_obs[key][:-1])

            for key in self.obs.keys():
                if len(self.obs[key].shape) == 6:
                    obs[key] = self.obs[key][:-1].transpose(1, 2, 0, 3, 4, 5).reshape(-1, *self.obs[key].shape[3:])
                elif len(self.obs[key].shape) == 5:
                    obs[key] = self.obs[key][:-1].transpose(1, 2, 0, 3, 4).reshape(-1, *self.obs[key].shape[3:])
                else:
                    obs[key] = _cast(self.obs[key][:-1])
        else:
            if len(self.share_obs.shape) > 4:
                share_obs = self.share_obs[:-1].transpose(1, 2, 0, 3, 4, 5).reshape(-1, *self.share_obs.shape[3:])
                obs = self.obs[:-1].transpose(1, 2, 0, 3, 4, 5).reshape(-1, *self.obs.shape[3:])
            else:
                share_obs = _cast(self.share_obs[:-1])
                obs = _cast(self.obs[:-1])

        actions = _cast(self.actions)
        action_log_probs = _cast(self.action_log_probs)
        advantages = _cast(advantages)
        value_preds = _cast(self.value_preds[:-1])
        returns = _cast(self.returns[:-1])
        masks = _cast(self.masks[:-1])
        active_masks = _cast(self.active_masks[:-1])
        # rnn_states = _cast(self.rnn_states[:-1])
        # rnn_states_critic = _cast(self.rnn_states_critic[:-1])
        rnn_states = self.rnn_states[:-1].transpose(1, 2, 0, 3, 4).reshape(-1, *self.rnn_states.shape[3:])
        rnn_states_critic = self.rnn_states_critic[:-1].transpose(1, 2, 0, 3, 4).reshape(-1,
                                                                                         *self.rnn_states_critic.shape[
                                                                                          3:])

        if self.available_actions is not None:
            available_actions = _cast(self.available_actions[:-1])

        for indices in sampler:

            if self._mixed_obs:
                share_obs_batch = defaultdict(list)
                obs_batch = defaultdict(list)
            else:
                share_obs_batch = []
                obs_batch = []

            rnn_states_batch = []
            rnn_states_critic_batch = []
            actions_batch = []
            available_actions_batch = []
            value_preds_batch = []
            return_batch = []
            masks_batch = []
            active_masks_batch = []
            old_action_log_probs_batch = []
            adv_targ = []

            for index in indices:

                ind = index * data_chunk_length
                # size [T+1 N M Dim]-->[T N M Dim]-->[N,M,T,Dim]-->[N*M*T,Dim]-->[L,Dim]
                if self._mixed_obs:
                    for key in share_obs.keys():
                        share_obs_batch[key].append(share_obs[key][ind:ind + data_chunk_length])
                    for key in obs.keys():
                        obs_batch[key].append(obs[key][ind:ind + data_chunk_length])
                else:
                    share_obs_batch.append(share_obs[ind:ind + data_chunk_length])
                    obs_batch.append(obs[ind:ind + data_chunk_length])

                actions_batch.append(actions[ind:ind + data_chunk_length])
                if self.available_actions is not None:
                    available_actions_batch.append(available_actions[ind:ind + data_chunk_length])
                value_preds_batch.append(value_preds[ind:ind + data_chunk_length])
                return_batch.append(returns[ind:ind + data_chunk_length])
                masks_batch.append(masks[ind:ind + data_chunk_length])
                active_masks_batch.append(active_masks[ind:ind + data_chunk_length])
                old_action_log_probs_batch.append(action_log_probs[ind:ind + data_chunk_length])
                adv_targ.append(advantages[ind:ind + data_chunk_length])
                # size [T+1 N M Dim]-->[T N M Dim]-->[N M T Dim]-->[N*M*T,Dim]-->[1,Dim]
                rnn_states_batch.append(rnn_states[ind])
                rnn_states_critic_batch.append(rnn_states_critic[ind])

            L, N = data_chunk_length, mini_batch_size

            # These are all from_numpys of size (L, N, Dim)
            if self._mixed_obs:
                for key in share_obs_batch.keys():
                    share_obs_batch[key] = np.stack(share_obs_batch[key], axis=1)
                for key in obs_batch.keys():
                    obs_batch[key] = np.stack(obs_batch[key], axis=1)
            else:
                share_obs_batch = np.stack(share_obs_batch, axis=1)
                obs_batch = np.stack(obs_batch, axis=1)

            actions_batch = np.stack(actions_batch, axis=1)
            if self.available_actions is not None:
                available_actions_batch = np.stack(available_actions_batch, axis=1)
            value_preds_batch = np.stack(value_preds_batch, axis=1)
            return_batch = np.stack(return_batch, axis=1)
            masks_batch = np.stack(masks_batch, axis=1)
            active_masks_batch = np.stack(active_masks_batch, axis=1)
            old_action_log_probs_batch = np.stack(old_action_log_probs_batch, axis=1)
            adv_targ = np.stack(adv_targ, axis=1)

            # States is just a (N, -1) from_numpy
            rnn_states_batch = np.stack(rnn_states_batch).reshape(N, *self.rnn_states.shape[3:])
            rnn_states_critic_batch = np.stack(rnn_states_critic_batch).reshape(N, *self.rnn_states_critic.shape[3:])

            # Flatten the (L, N, ...) from_numpys to (L * N, ...)
            if self._mixed_obs:
                for key in share_obs_batch.keys():
                    share_obs_batch[key] = _flatten(L, N, share_obs_batch[key])
                for key in obs_batch.keys():
                    obs_batch[key] = _flatten(L, N, obs_batch[key])
            else:
                share_obs_batch = _flatten(L, N, share_obs_batch)
                obs_batch = _flatten(L, N, obs_batch)
            actions_batch = _flatten(L, N, actions_batch)
            if self.available_actions is not None:
                available_actions_batch = _flatten(L, N, available_actions_batch)
            else:
                available_actions_batch = None
            value_preds_batch = _flatten(L, N, value_preds_batch)
            return_batch = _flatten(L, N, return_batch)
            masks_batch = _flatten(L, N, masks_batch)
            active_masks_batch = _flatten(L, N, active_masks_batch)
            old_action_log_probs_batch = _flatten(L, N, old_action_log_probs_batch)
            adv_targ = _flatten(L, N, adv_targ)

            yield share_obs_batch, obs_batch, rnn_states_batch, rnn_states_critic_batch, actions_batch, value_preds_batch, return_batch, masks_batch, active_masks_batch, old_action_log_probs_batch, adv_targ, available_actions_batch



import copy


import torch
import torch.nn as nn

def _t2n(x):
    return x.detach().cpu().numpy()


class FixedCategorical(torch.distributions.Categorical):
    def sample(self):
        return super().sample().unsqueeze(-1)

    def log_probs(self, actions):
        return (
            super()
            .log_prob(actions.squeeze(-1))
            .view(actions.size(0), -1)
            .sum(-1)
            .unsqueeze(-1)
        )

    def mode(self):
        return self.probs.argmax(dim=-1, keepdim=True)


# Normal
class FixedNormal(torch.distributions.Normal):
    def log_probs(self, actions):
        return super().log_prob(actions).sum(-1, keepdim=True)

    def entrop(self):
        return super.entropy().sum(-1)

    def mode(self):
        return self.mean


# Bernoulli
class FixedBernoulli(torch.distributions.Bernoulli):
    def log_probs(self, actions):
        return super.log_prob(actions).view(actions.size(0), -1).sum(-1).unsqueeze(-1)

    def entropy(self):
        return super().entropy().sum(-1)

    def mode(self):
        return torch.gt(self.probs, 0.5).float()


class Categorical(nn.Module):
    def __init__(self, num_inputs, num_outputs, use_orthogonal=True, gain=0.01):
        super(Categorical, self).__init__()
        init_method = [nn.init.xavier_uniform_, nn.init.orthogonal_][use_orthogonal]

        def init_(m):
            return init(m, init_method, lambda x: nn.init.constant_(x, 0), gain)

        self.linear = init_(nn.Linear(num_inputs, num_outputs))

    def forward(self, x, available_actions=None):
        x = self.linear(x)
        if available_actions is not None:
            x[available_actions == 0] = -1e10
        return FixedCategorical(logits=x)


class DiagGaussian(nn.Module):
    def __init__(self, num_inputs, num_outputs, use_orthogonal=True, gain=0.01):
        super(DiagGaussian, self).__init__()

        init_method = [nn.init.xavier_uniform_, nn.init.orthogonal_][use_orthogonal]

        def init_(m):
            return init(m, init_method, lambda x: nn.init.constant_(x, 0), gain)

        self.fc_mean = init_(nn.Linear(num_inputs, num_outputs))
        self.logstd = AddBias(torch.zeros(num_outputs))

    def forward(self, x):
        action_mean = self.fc_mean(x)

        #  An ugly hack for my KFAC implementation.
        zeros = torch.zeros(action_mean.size())
        if x.is_cuda:
            zeros = zeros.cuda()

        action_logstd = self.logstd(zeros)
        return FixedNormal(action_mean, action_logstd.exp())


class Bernoulli(nn.Module):
    def __init__(self, num_inputs, num_outputs, use_orthogonal=True, gain=0.01):
        super(Bernoulli, self).__init__()
        init_method = [nn.init.xavier_uniform_, nn.init.orthogonal_][use_orthogonal]

        def init_(m):
            return init(m, init_method, lambda x: nn.init.constant_(x, 0), gain)

        self.linear = init_(nn.Linear(num_inputs, num_outputs))

    def forward(self, x):
        x = self.linear(x)
        return FixedBernoulli(logits=x)


class AddBias(nn.Module):
    def __init__(self, bias):
        super(AddBias, self).__init__()
        self._bias = nn.Parameter(bias.unsqueeze(1))

    def forward(self, x):
        if x.dim() == 2:
            bias = self._bias.t().view(1, -1)
        else:
            bias = self._bias.t().view(1, -1, 1, 1)

        return x + bias


def init(module, weight_init, bias_init, gain=1):
    weight_init(module.weight.data, gain=gain)
    if module.bias is not None:
        bias_init(module.bias.data)
    return module

def get_clones(module, N):
    return nn.ModuleList([copy.deepcopy(module) for i in range(N)])


class MLPLayer(nn.Module):
    def __init__(self, input_dim, hidden_size, layer_N, use_orthogonal, activation_id):
        super(MLPLayer, self).__init__()
        self._layer_N = layer_N

        active_func = [nn.Tanh(), nn.ReLU(), nn.LeakyReLU(), nn.ELU()][activation_id]
        init_method = [nn.init.xavier_uniform_, nn.init.orthogonal_][use_orthogonal]
        gain = nn.init.calculate_gain(['tanh', 'relu', 'leaky_relu', 'leaky_relu'][activation_id])

        def init_(m):
            return init(m, init_method, lambda x: nn.init.constant_(x, 0), gain=gain)

        self.fc1 = nn.Sequential(
            init_(nn.Linear(input_dim, hidden_size)), active_func, nn.LayerNorm(hidden_size))
        self.fc_h = nn.Sequential(init_(
            nn.Linear(hidden_size, hidden_size)), active_func, nn.LayerNorm(hidden_size))
        self.fc2 = get_clones(self.fc_h, self._layer_N)

    def forward(self, x):
        x = self.fc1(x)
        for i in range(self._layer_N):
            x = self.fc2[i](x)
        return x

class MLPBase2(nn.Module):
    def __init__(self, args, obs_dim, use_attn_internal=False):
        super(MLPBase2, self).__init__()

        self._use_feature_normalization = args.use_feature_normalization
        self._use_orthogonal = args.use_orthogonal
        self._activation_id = args.activation_id
        self._use_attn = args.use_attn
        self._use_attn_internal = use_attn_internal
        self._use_average_pool = args.use_average_pool
        self._use_conv1d = args.use_conv1d
        self._stacked_frames = args.stacked_frames
        self._layer_N = 0 if args.use_single_network else args.layer_N
        self._attn_size = args.attn_size
        self.hidden_size = args.hidden_size



        if self._use_feature_normalization:
            self.feature_norm = nn.LayerNorm(obs_dim)


        inputs_dim = obs_dim


        self.mlp = MLPLayer(inputs_dim, self.hidden_size,
                            self._layer_N, self._use_orthogonal, self._activation_id)

    def forward(self, x):

        if self._use_feature_normalization:
            x = self.feature_norm(x)

        x = self.mlp(x)

        return x

    @property
    def output_size(self):
        return self.hidden_size


class RNNLayer2(nn.Module):
    def __init__(self, inputs_dim, outputs_dim, recurrent_N, use_orthogonal,rnn_type='gru'):
        super(RNNLayer2, self).__init__()
        self._recurrent_N = recurrent_N
        self._use_orthogonal = use_orthogonal
        self.rnn_type = rnn_type
        if rnn_type == 'gru':
            self.rnn = nn.GRU(inputs_dim, outputs_dim, num_layers=self._recurrent_N)
        elif rnn_type == 'lstm':
            self.rnn = nn.LSTM(inputs_dim, outputs_dim, num_layers=self._recurrent_N)
        else:
            raise NotImplementedError(f'RNN type {rnn_type} has not been implemented.')

        for name, param in self.rnn.named_parameters():
            if 'bias' in name:
                nn.init.constant_(param, 0)
            elif 'weight' in name:
                if self._use_orthogonal:
                    nn.init.orthogonal_(param)
                else:
                    nn.init.xavier_uniform_(param)
        self.norm = nn.LayerNorm(outputs_dim)

    def rnn_forward(self, x, h):
        if self.rnn_type == 'lstm':
            h = torch.split(h, h.shape[-1] // 2, dim=-1)
            h = (h[0].contiguous(), h[1].contiguous())
        x_, h_ = self.rnn(x, h)
        if self.rnn_type == 'lstm':
            h_ = torch.cat(h_, -1)
        return x_, h_

    def forward(self, x, hxs, masks):
        if x.size(0) == hxs.size(0):
            x, hxs = self.rnn_forward(x.unsqueeze(0), (hxs * masks.repeat(1, self._recurrent_N).unsqueeze(-1)).transpose(0, 1).contiguous())
            #x= self.gru(x.unsqueeze(0))
            x = x.squeeze(0)
            hxs = hxs.transpose(0, 1)
        else:
            # x is a (T, N, -1) tensor that has been flatten to (T * N, -1)
            N = hxs.size(0)
            T = int(x.size(0) / N)

            # unflatten
            x = x.view(T, N, x.size(1))

            # Same deal with masks
            masks = masks.view(T, N)

            # Let's figure out which steps in the sequence have a zero for any agent
            # We will always assume t=0 has a zero in it as that makes the logic cleaner
            has_zeros = ((masks[1:] == 0.0)
                         .any(dim=-1)
                         .nonzero()
                         .squeeze()
                         .cpu())

            # +1 to correct the masks[1:]
            if has_zeros.dim() == 0:
                # Deal with scalar
                has_zeros = [has_zeros.item() + 1]
            else:
                has_zeros = (has_zeros + 1).numpy().tolist()

            # add t=0 and t=T to the list
            has_zeros = [0] + has_zeros + [T]

            hxs = hxs.transpose(0, 1)

            outputs = []
            for i in range(len(has_zeros) - 1):
                # We can now process steps that don't have any zeros in masks together!
                # This is much faster
                start_idx = has_zeros[i]
                end_idx = has_zeros[i + 1]
                temp = (hxs * masks[start_idx].view(1, -1, 1).repeat(self._recurrent_N, 1, 1)).contiguous()
                rnn_scores, hxs = self.rnn_forward(x[start_idx:end_idx], temp)
                outputs.append(rnn_scores)

            # assert len(outputs) == T
            # x is a (T, N, -1) tensor
            x = torch.cat(outputs, dim=0)

            # flatten
            x = x.reshape(T * N, -1)
            hxs = hxs.transpose(0, 1)

        x = self.norm(x)
        return x, hxs


class ACTLayer2(nn.Module):
    def __init__(self, action_dim, inputs_dim, use_orthogonal, gain):
        super(ACTLayer2, self).__init__()
        self.multidiscrete_action = False
        self.continuous_action = False
        self.mixed_action = False


        action_dim = action_dim
        self.action_out = Categorical(inputs_dim, action_dim, use_orthogonal, gain)


    def forward(self, x, available_actions=None, deterministic=False):

        action_logits = self.action_out(x, available_actions)
        actions = action_logits.mode() if deterministic else action_logits.sample()
        action_probs = action_logits.probs

        return actions, action_probs

    def get_probs(self, x, available_actions=None):
        if self.mixed_action or self.multidiscrete_action:
            action_probs = []
            for action_out in self.action_outs:
                action_logit = action_out(x)
                action_prob = action_logit.probs
                action_probs.append(action_prob)
            action_probs = torch.cat(action_probs, -1)
        elif self.continuous_action:
            action_logits = self.action_out(x)
            action_probs = action_logits.probs
        else:
            action_logits = self.action_out(x, available_actions)
            action_probs = action_logits.probs

        return action_probs

    def evaluate_actions(self, x, action, available_actions=None, active_masks=None):
        if self.mixed_action:
            a, b = action.split((2, 1), -1)
            b = b.long()
            action = [a, b]
            action_log_probs = []
            dist_entropy = []
            for action_out, act in zip(self.action_outs, action):
                action_logit = action_out(x)
                action_log_probs.append(action_logit.log_probs(act))
                if active_masks is not None:
                    if len(action_logit.entropy().shape) == len(active_masks.shape):
                        dist_entropy.append((action_logit.entropy() * active_masks).sum() / active_masks.sum())
                    else:
                        dist_entropy.append(
                            (action_logit.entropy() * active_masks.squeeze(-1)).sum() / active_masks.sum())
                else:
                    dist_entropy.append(action_logit.entropy().mean())

            action_log_probs = torch.sum(torch.cat(action_log_probs, -1), -1, keepdim=True)
            dist_entropy = dist_entropy[0] * 0.0025 + dist_entropy[1] * 0.01

        elif self.multidiscrete_action:
            action = torch.transpose(action, 0, 1)
            action_log_probs = []
            dist_entropy = []
            for action_out, act in zip(self.action_outs, action):
                action_logit = action_out(x)
                action_log_probs.append(action_logit.log_probs(act))
                if active_masks is not None:
                    dist_entropy.append((action_logit.entropy() * active_masks.squeeze(-1)).sum() / active_masks.sum())
                else:
                    dist_entropy.append(action_logit.entropy().mean())

            action_log_probs = torch.cat(action_log_probs, -1)  # ! could be wrong
            dist_entropy = torch.tensor(dist_entropy).mean()

        elif self.continuous_action:
            action_logits = self.action_out(x)
            action_log_probs = action_logits.log_probs(action)
            if active_masks is not None:
                dist_entropy = (action_logits.entropy() * active_masks).sum() / active_masks.sum()
            else:
                dist_entropy = action_logits.entropy().mean()
        else:
            action_logits = self.action_out(x, available_actions)
            action_log_probs = action_logits.log_probs(action)
            if active_masks is not None:
                dist_entropy = (action_logits.entropy() * active_masks.squeeze(-1)).sum() / active_masks.sum()
            else:
                dist_entropy = action_logits.entropy().mean()

        return action_log_probs, dist_entropy


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2023 The TARTRL Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""""""
import io
import pathlib

from typing import Union, Dict, Tuple, Optional

import numpy as np
import torch
import gym

from tartrl.runners.common.base_agent import BaseAgent
from tartrl.runners.common.base_agent import SelfAgent

from tartrl.drivers.onpolicy_driver import OnPolicyDriver as Driver
from tartrl.algorithms.ppo import PPOAlgorithm as TrainAlgo
from tartrl.buffers import NormalReplayBuffer as ReplayBuffer
from tartrl.utils.logger import Logger

class PPOAgent(BaseAgent):
    def __init__(
            self,
            net: Optional[torch.nn.Module] = None,
            env: Union[gym.Env, str] = None,
            run_dir: str = "./",
            n_rollout_threads: Optional[int] = None,
            rank: int = 0,
            world_size: int = 1,
            agent_num: int = 1,
            use_wandb: bool = False,
            use_tensorboard: bool = False,
    ) -> None:
        self.net = net
        self._args = net.args
        self._use_wandb = use_wandb
        self._use_tensorboard = not use_wandb and use_tensorboard

        if n_rollout_threads is not None:
            self._args.n_rollout_threads = n_rollout_threads

        if env is not None:
            self._env = env
        elif hasattr(net, "env") and net.env is not None:
            self._env = net.env
        else:
            raise ValueError("env is None")

        self.rank = rank
        self.world_size = world_size

        self.client = None
        self.agent_num = agent_num
        self.run_dir = run_dir

    def train(
            self: SelfAgent,
            total_time_steps: int
    ) -> None:

        self._args.num_env_steps = total_time_steps

        self.config = {
            "args": self._args,
            "num_agents": self.agent_num,
            "run_dir": self.run_dir,
            'envs': self._env,
            'device': self.net.device,
        }

        trainer = TrainAlgo(
            args=self._args,
            init_module=self.net.module,
            device=self.net.device)

        share_observation_space = self._env.observation_space

        buffer = ReplayBuffer(
            self._args,
            self.agent_num,
            self._env.observation_space,
            share_observation_space,
            self._env.action_space,
            data_client=None,
        )

        logger = Logger(
            args = self._args,
            project_name="PPOAgent",
            scenario_name=self._env.env_name,
            wandb_entity=self._args.wandb_entity,
            exp_name="ppo",
            log_path="../../../ppo_agent_logger/",
            use_wandb=self._use_wandb,
            use_tensorboard=self._use_tensorboard,
        )
        driver = Driver(
            config=self.config,
            trainer=trainer,
            buffer=buffer,
            client=self.client,
            rank=self.rank,
            world_size=self.world_size,
            logger = logger,
        )
        driver.run()

    def act(
            self,
            observation: Union[np.ndarray, Dict[str, np.ndarray]],
            deterministic: bool = False,
    ) -> Tuple[np.ndarray, Optional[Tuple[np.ndarray, ...]]]:
        assert self.net is not None, "net is None"
        return self.net.act(observation, deterministic=deterministic)


    def reset(self) -> None:
        self.net.reset()

    def save(self, path: Union[str, pathlib.Path, io.BufferedIOBase]) -> None:
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2021 The TARTRL Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""""""

import torch
import torch.nn as nn

from tartrl.modules.networks.utils.util import init
from tartrl.modules.networks.utils.mlp import MLPBase, MLPLayer
from tartrl.modules.networks.utils.rnn import RNNLayer
from tartrl.modules.networks.utils.popart import PopArt
from tartrl.modules.networks.utils.mix import MIXBase
from tartrl.modules.networks.utils.cnn import CNNBase
from tartrl.modules.networks.utils.FcEncoder import FcEncoder
from tartrl.modules.networks.base_value_network import BaseValueNetwork

from tartrl.buffers.utils.util import get_shape_from_obs_space
from .utils.utils import make_models_for_obs, RecurrentBackbone, PopArtValueHead, recursive_apply
from tartrl.utils.util import check_v2 as check


class InputEncoder(nn.Module):
    def __init__(self):
        super(InputEncoder, self).__init__()
        fc_layer_num = 2
        fc_output_num = 64

        self.ball_owner_input_num = 35
        self.left_input_num = 88
        self.right_input_num = 88
        self.match_state_input_num = 9

        self.ball_owner_encoder = FcEncoder(fc_layer_num, self.ball_owner_input_num, fc_output_num)
        self.left_encoder = FcEncoder(fc_layer_num, self.left_input_num, fc_output_num)
        self.right_encoder = FcEncoder(fc_layer_num, self.right_input_num, fc_output_num)
        self.match_state_encoder = FcEncoder(fc_layer_num, self.match_state_input_num, self.match_state_input_num)

    def forward(self, x):
        ball_owner_vec = x[:, :self.ball_owner_input_num]
        left_vec = x[:, self.ball_owner_input_num: self.ball_owner_input_num + self.left_input_num]
        right_vec = x[:,
                    self.ball_owner_input_num + self.left_input_num:  self.ball_owner_input_num + self.left_input_num + self.right_input_num]
        match_state_vec = x[:, self.ball_owner_input_num + self.left_input_num + self.right_input_num:]

        ball_owner_output = self.ball_owner_encoder(ball_owner_vec)
        left_output = self.left_encoder(left_vec)
        right_output = self.right_encoder(right_vec)
        match_state_output = self.match_state_encoder(match_state_vec)

        return torch.cat([
            ball_owner_output,
            left_output,
            right_output,
            match_state_output
        ], 1)


class NewValueNetwork(BaseValueNetwork):
    def __init__(self, args, share_obs_space, device=torch.device("cpu")):
        super(NewValueNetwork, self).__init__(args, device)
        self.hidden_size = args.hidden_size
        self.tpdv = dict(dtype=torch.float32, device=device)

        input_embedding_size = 64 * 3 + 9
        self.input_encoder = InputEncoder()
        self.base = FcEncoder(3, input_embedding_size, self.hidden_size)
        self.rnn = RNNLayer(self.hidden_size, self.hidden_size, args.recurrent_N, args.use_orthogonal,
                            rnn_type=args.rnn_type)
        self.v_out = nn.Linear(self.hidden_size, 1)

        self.to(device)

    def forward(self, share_obs, rnn_states, masks):
        share_obs = check(share_obs).to(**self.tpdv)
        rnn_states = check(rnn_states).to(**self.tpdv)
        masks = check(masks).to(**self.tpdv)

        critic_features = self.input_encoder(share_obs)
        critic_features = self.base(critic_features)

        critic_features, rnn_states = self.rnn(critic_features, rnn_states, masks)
        values = self.v_out(critic_features)
        return values, rnn_states


class ValueNetwork(BaseValueNetwork):
    def __init__(self, args, input_space, action_space=None, use_half=False, device=torch.device("cpu")):
        super(ValueNetwork, self).__init__(args, device)

        self.hidden_size = args.hidden_size
        self._use_orthogonal = args.use_orthogonal
        self._activation_id = args.activation_id
        self._use_naive_recurrent_policy = args.use_naive_recurrent_policy
        self._use_recurrent_policy = args.use_recurrent_policy
        self._use_influence_policy = args.use_influence_policy
        self._use_popart = args.use_popart
        self._influence_layer_N = args.influence_layer_N
        self._recurrent_N = args.recurrent_N
        self.tpdv = dict(dtype=torch.float32, device=device)

        init_method = [nn.init.xavier_uniform_, nn.init.orthogonal_][self._use_orthogonal]

        share_obs_shape = get_shape_from_obs_space(input_space)

        if 'Dict' in share_obs_shape.__class__.__name__:
            self._mixed_obs = True
            self.base = MIXBase(args, share_obs_shape, cnn_layers_params=args.cnn_layers_params)
        else:
            self._mixed_obs = False
            self.base = CNNBase(args, share_obs_shape) if len(share_obs_shape) == 3 else MLPBase(args, share_obs_shape,
                                                                                                 use_attn_internal=True,
                                                                                                 use_cat_self=args.use_cat_self)

        input_size = self.base.output_size

        if self._use_naive_recurrent_policy or self._use_recurrent_policy:
            self.rnn = RNNLayer(input_size, self.hidden_size, self._recurrent_N, self._use_orthogonal,
                                rnn_type=args.rnn_type)
            input_size = self.hidden_size

        if self._use_influence_policy:
            self.mlp = MLPLayer(share_obs_shape[0], self.hidden_size,
                                self._influence_layer_N, self._use_orthogonal, self._activation_id)
            input_size += self.hidden_size

        def init_(m):
            return init(m, init_method, lambda x: nn.init.constant_(x, 0))

        if self._use_popart:
            self.v_out = init_(PopArt(input_size, 1, device=device))
        else:
            self.v_out = init_(nn.Linear(input_size, 1))

        self.to(device)

    def forward(self, share_obs, rnn_states, masks):
        if self._mixed_obs:
            for key in share_obs.keys():
                share_obs[key] = check(share_obs[key]).to(**self.tpdv)
        else:
            share_obs = check(share_obs).to(**self.tpdv)
        rnn_states = check(rnn_states).to(**self.tpdv)
        masks = check(masks).to(**self.tpdv)

        critic_features = self.base(share_obs)

        if self._use_naive_recurrent_policy or self._use_recurrent_policy:
            critic_features, rnn_states = self.rnn(critic_features, rnn_states, masks)

        if self._use_influence_policy:
            mlp_share_obs = self.mlp(share_obs)
            critic_features = torch.cat([critic_features, mlp_share_obs], dim=1)

        values = self.v_out(critic_features)

        return values, rnn_states


class XZLValueNetwork(BaseValueNetwork):
    def __init__(self, args, share_obs_space, device=torch.device("cpu")):
        super(XZLValueNetwork, self).__init__(args, device)
        self.tpdv = dict(dtype=torch.float32, device=device)
        share_obs_shape = get_shape_from_obs_space(share_obs_space)
        obs_dim = share_obs_shape[0]
        hidden_dim = args.hidden_size
        cnn_layers = None
        activation = ['tanh', 'relu', 'leaky_relu', 'leaky_relu'][args.activation_id]
        dense_layers = args.layer_N
        num_rnn_layers = args.rnn_num
        rnn_type = args.rnn_type
        layernorm = True

        if activation == 'relu':
            self.activation = nn.ReLU
        elif activation == 'tanh':
            self.activation = nn.Tanh
        else:
            raise NotImplementedError(f"Activation function {activation} not implemented.")

        if isinstance(obs_dim, int):
            obs_dim = {"obs": obs_dim}

        self.state_modules_dict = make_models_for_obs(obs_dim=obs_dim,
                                                      hidden_dim=hidden_dim,
                                                      cnn_layers=cnn_layers,
                                                      activation=self.activation)

        self.critic_backbone = RecurrentBackbone(obs_dim=hidden_dim * len(obs_dim),
                                                 hidden_dim=hidden_dim,
                                                 dense_layers=dense_layers,
                                                 num_rnn_layers=num_rnn_layers,
                                                 rnn_type=rnn_type,
                                                 activation=activation,
                                                 layernorm=layernorm,
                                                 batch_first=False)
        f_ = self.critic_backbone.feature_dim
        value_dim = 1
        self.critic_head = PopArtValueHead(f_, value_dim)
        self.to(device)

    def forward(self, share_obs, rnn_states, masks):

        share_obs = check(share_obs).to(**self.tpdv)
        rnn_states = check(rnn_states).to(**self.tpdv)
        masks = check(masks).to(**self.tpdv)

        obs_ = dict(obs=share_obs)

        on_reset = masks
        state = torch.cat([m(obs_[k]) for k, m in self.state_modules_dict.items()], dim=-1)
        critic_features, critic_hx = self.critic_backbone(state, rnn_states, on_reset)

        critic = self.critic_head(critic_features)
        values = critic.squeeze(0)
        rnn_states = recursive_apply(critic_hx, lambda x: x.transpose(0, 1))

        return values, rnn_states




import numpy as np

import torch
import torch.nn as nn
import torch.nn.functional as F

class RNNLayer(nn.Module):
    def __init__(self, inputs_dim, outputs_dim, recurrent_N, use_orthogonal,rnn_type='gru'):
        super(RNNLayer, self).__init__()
        self._recurrent_N = recurrent_N
        self._use_orthogonal = use_orthogonal
        self.rnn_type = rnn_type
        if rnn_type == 'gru':
            self.rnn = nn.GRU(inputs_dim, outputs_dim, num_layers=self._recurrent_N)
        elif rnn_type == 'lstm':
            self.rnn = nn.LSTM(inputs_dim, outputs_dim, num_layers=self._recurrent_N)
        else:
            raise NotImplementedError(f'RNN type {rnn_type} has not been implemented.')

        for name, param in self.rnn.named_parameters():
            if 'bias' in name:
                nn.init.constant_(param, 0)
            elif 'weight' in name:
                if self._use_orthogonal:
                    nn.init.orthogonal_(param)
                else:
                    nn.init.xavier_uniform_(param)
        self.norm = nn.LayerNorm(outputs_dim)

    def rnn_forward(self, x, h):
        if self.rnn_type == 'lstm':
            h = torch.split(h, h.shape[-1] // 2, dim=-1)
            h = (h[0].contiguous(), h[1].contiguous())
        x_, h_ = self.rnn(x, h)
        if self.rnn_type == 'lstm':
            h_ = torch.cat(h_, -1)
        return x_, h_

    def forward(self, x, hxs, masks):
        if x.size(0) == hxs.size(0):
            x, hxs = self.rnn_forward(x.unsqueeze(0), (hxs * masks.repeat(1, self._recurrent_N).unsqueeze(-1)).transpose(0, 1).contiguous())
            #x= self.gru(x.unsqueeze(0))
            x = x.squeeze(0)
            hxs = hxs.transpose(0, 1)
        else:
            # x is a (T, N, -1) tensor that has been flatten to (T * N, -1)
            N = hxs.size(0)
            T = int(x.size(0) / N)

            # unflatten
            x = x.view(T, N, x.size(1))

            # Same deal with masks
            masks = masks.view(T, N)

            # Let's figure out which steps in the sequence have a zero for any agent
            # We will always assume t=0 has a zero in it as that makes the logic cleaner
            has_zeros = ((masks[1:] == 0.0)
                         .any(dim=-1)
                         .nonzero()
                         .squeeze()
                         .cpu())

            # +1 to correct the masks[1:]
            if has_zeros.dim() == 0:
                # Deal with scalar
                has_zeros = [has_zeros.item() + 1]
            else:
                has_zeros = (has_zeros + 1).numpy().tolist()

            # add t=0 and t=T to the list
            has_zeros = [0] + has_zeros + [T]

            hxs = hxs.transpose(0, 1)

            outputs = []
            for i in range(len(has_zeros) - 1):
                # We can now process steps that don't have any zeros in masks together!
                # This is much faster
                start_idx = has_zeros[i]
                end_idx = has_zeros[i + 1]               
                temp = (hxs * masks[start_idx].view(1, -1, 1).repeat(self._recurrent_N, 1, 1)).contiguous()
                rnn_scores, hxs = self.rnn_forward(x[start_idx:end_idx], temp)
                outputs.append(rnn_scores)

            # assert len(outputs) == T
            # x is a (T, N, -1) tensor
            x = torch.cat(outputs, dim=0)

            # flatten
            x = x.reshape(T * N, -1)
            hxs = hxs.transpose(0, 1)

        x = self.norm(x)
        return x, hxs



import math

def get_gard_norm(it):
    sum_grad = 0
    for x in it:
        if x.grad is None:
            continue
        sum_grad += x.grad.norm() ** 2
    return math.sqrt(sum_grad)


def update_linear_schedule(optimizer, epoch, total_num_epochs, initial_lr):
    """Decreases the learning rate linearly"""
    lr = initial_lr - (initial_lr * (epoch / float(total_num_epochs)))
    for param_group in optimizer.param_groups:
        param_group['lr'] = lr


def huber_loss(e, d):
    a = (abs(e) <= d).float()
    b = (e > d).float()
    return a*e**2/2 + b*d*(abs(e)-d/2)


def mse_loss(e):
    return e**2/2




from .distributions import Bernoulli, Categorical, DiagGaussian
import math
import numpy as np

import torch
import torch.nn as nn
import torch.nn.functional as F

class ACTLayer(nn.Module):
    def __init__(self, action_space, inputs_dim, use_orthogonal, gain):
        super(ACTLayer, self).__init__()
        self.multidiscrete_action = False
        self.continuous_action = False
        self.mixed_action = False

        if action_space.__class__.__name__ == "Discrete":
            action_dim = action_space.n
            self.action_out = Categorical(inputs_dim, action_dim, use_orthogonal, gain)

        elif action_space.__class__.__name__ == "Box":
            self.continuous_action = True
            action_dim = action_space.shape[0]
            self.action_out = DiagGaussian(inputs_dim, action_dim, use_orthogonal, gain)
        elif action_space.__class__.__name__ == "MultiBinary":
            action_dim = action_space.shape[0]
            self.action_out = Bernoulli(inputs_dim, action_dim, use_orthogonal, gain)
        elif action_space.__class__.__name__ == "MultiDiscrete":
            self.multidiscrete_action = True
            action_dims = action_space.high - action_space.low + 1
            self.action_outs = []
            for action_dim in action_dims:
                self.action_outs.append(Categorical(inputs_dim, action_dim, use_orthogonal, gain))
            self.action_outs = nn.ModuleList(self.action_outs)
        else:  # discrete + continous
            self.mixed_action = True
            continous_dim = action_space[0].shape[0]
            discrete_dim = action_space[1].n
            self.action_outs = nn.ModuleList([DiagGaussian(inputs_dim, continous_dim, use_orthogonal, gain), Categorical(
                inputs_dim, discrete_dim, use_orthogonal, gain)])
    
    def forward(self, x, available_actions=None, deterministic=False):
        if self.mixed_action :
            actions = []
            action_log_probs = []
            for action_out in self.action_outs:
                action_logit = action_out(x)
                action = action_logit.mode() if deterministic else action_logit.sample()
                action_log_prob = action_logit.log_probs(action)
                actions.append(action.float())
                action_log_probs.append(action_log_prob)

            actions = torch.cat(actions, -1)
            action_log_probs = torch.sum(torch.cat(action_log_probs, -1), -1, keepdim=True)

        elif self.multidiscrete_action:
            actions = []
            action_log_probs = []
            for action_out in self.action_outs:
                action_logit = action_out(x)
                action = action_logit.mode() if deterministic else action_logit.sample()
                action_log_prob = action_logit.log_probs(action)
                actions.append(action)
                action_log_probs.append(action_log_prob)

            actions = torch.cat(actions, -1)
            action_log_probs = torch.cat(action_log_probs, -1)
        
        elif self.continuous_action:
            action_logits = self.action_out(x)
            actions = action_logits.mode() if deterministic else action_logits.sample() 
            action_log_probs = action_logits.log_probs(actions)
        
        else:
            action_logits = self.action_out(x, available_actions)
            actions = action_logits.mode() if deterministic else action_logits.sample() 
            action_log_probs = action_logits.log_probs(actions)
        
        return actions, action_log_probs

    def get_probs(self, x, available_actions=None):
        if self.mixed_action or self.multidiscrete_action:
            action_probs = []
            for action_out in self.action_outs:
                action_logit = action_out(x)
                action_prob = action_logit.probs
                action_probs.append(action_prob)
            action_probs = torch.cat(action_probs, -1)
        elif self.continuous_action:
            action_logits = self.action_out(x)
            action_probs = action_logits.probs
        else:
            action_logits = self.action_out(x, available_actions)
            action_probs = action_logits.probs
        
        return action_probs

    def evaluate_actions(self, x, action, available_actions=None, active_masks=None, get_probs=False):
        if self.mixed_action:
            a, b = action.split((2, 1), -1)
            b = b.long()
            action = [a, b] 
            action_log_probs = [] 
            dist_entropy = []
            for action_out, act in zip(self.action_outs, action):
                action_logit = action_out(x)
                action_log_probs.append(action_logit.log_probs(act))
                if active_masks is not None:
                    if len(action_logit.entropy().shape) == len(active_masks.shape):
                        dist_entropy.append((action_logit.entropy() * active_masks).sum()/active_masks.sum()) 
                    else:
                        dist_entropy.append((action_logit.entropy() * active_masks.squeeze(-1)).sum()/active_masks.sum())
                else:
                    dist_entropy.append(action_logit.entropy().mean())
                
            action_log_probs = torch.sum(torch.cat(action_log_probs, -1), -1, keepdim=True)
            dist_entropy = dist_entropy[0] * 0.0025 + dist_entropy[1] * 0.01 

        elif self.multidiscrete_action:
            action = torch.transpose(action, 0, 1)
            action_log_probs = []
            dist_entropy = []
            for action_out, act in zip(self.action_outs, action):
                action_logit = action_out(x)
                action_log_probs.append(action_logit.log_probs(act))
                if active_masks is not None:
                    dist_entropy.append((action_logit.entropy()*active_masks.squeeze(-1)).sum()/active_masks.sum())
                else:
                    dist_entropy.append(action_logit.entropy().mean())

            action_log_probs = torch.cat(action_log_probs, -1) # ! could be wrong
            dist_entropy = torch.tensor(dist_entropy).mean()

        elif self.continuous_action:
            action_logits = self.action_out(x)
            action_log_probs = action_logits.log_probs(action)
            act_entropy = action_logits.entropy()
            # import pdb;pdb.set_trace()
            if active_masks is not None:
                dist_entropy = (act_entropy*active_masks).sum()/active_masks.sum()
            else:
                dist_entropy = act_entropy.mean()

        else:
            action_logits = self.action_out(x, available_actions)
            action_log_probs = action_logits.log_probs(action)
            if active_masks is not None:
                dist_entropy = (action_logits.entropy()*active_masks.squeeze(-1)).sum()/active_masks.sum()
            else:
                dist_entropy = action_logits.entropy().mean()
        if not get_probs:
            return action_log_probs, dist_entropy
        else:
            return action_log_probs, dist_entropy, action_logits


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2023 The TARTRL Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""""""
from typing import Dict, Any, Optional
import numpy as np
import torch
from torch.nn.parallel import DistributedDataParallel

from tartrl.drivers.rl_driver import RLDriver

from tartrl.utils.util import _t2n
from tartrl.utils.logger import Logger
from tartrl.envs.wrappers.venv import VecInfo

class OnPolicyDriver(RLDriver):
    def __init__(
            self,
            config: Dict[str, Any],
            trainer,
            buffer,
            rank: int = 0,
            world_size: int = 1,
            client=None,
            logger: Optional[Logger] = None,
    ) -> None:

        self.trainer = trainer
        self.buffer = buffer
        self.learner_episode = -1
        self.actor_id = 0
        self.weight_ids = [0]
        self.world_size = world_size
        self.logger = logger

        args = config['args']
        self.program_type = args.program_type

        self.envs = config['envs']
        self.device = config['device']

        assert not (
                self.program_type != "actor" and self.world_size is None), "world size can not be none, get {}".format(
            world_size)

        self.num_agents = config['num_agents']

        # print(args)
        assert isinstance(
            rank, int), "rank must be int, but get {}".format(rank)
        self.rank = rank

        # for distributed learning
        assert not (world_size is None and self.program_type == 'learner'), "world_size must be int, but get {}".format(
            world_size)

        # parameters
        self.env_name = args.env_name
        self.algorithm_name = args.algorithm_name
        self.experiment_name = args.experiment_name
        self.concat_obs_as_share_obs = args.concat_obs_as_share_obs
        self.use_obs_instead_of_state = args.use_obs_instead_of_state
        self.num_env_steps = args.num_env_steps
        self.episode_length = args.episode_length
        self.n_rollout_threads = args.n_rollout_threads
        self.learner_n_rollout_threads = args.learner_n_rollout_threads

        self.n_eval_rollout_threads = args.n_eval_rollout_threads
        self.n_render_rollout_threads = args.n_render_rollout_threads
        self.use_linear_lr_decay = args.use_linear_lr_decay
        self.hidden_size = args.hidden_size
        self.use_wandb = not args.disable_wandb
        self.use_single_network = args.use_single_network
        self.use_render = args.use_render
        self.use_transmit = args.use_transmit
        self.recurrent_N = args.recurrent_N

        self.only_eval = args.only_eval
        # interval
        self.save_interval = args.save_interval
        self.use_eval = args.use_eval
        self.eval_interval = args.eval_interval
        self.log_interval = args.log_interval

        self.distributed_type = args.distributed_type

        self.actor_num = args.actor_num

        if self.distributed_type == 'async' and self.program_type == 'whole':
            print('can\'t use async mode when program_type is whole!')
            exit()

        if self.program_type in ["whole", "local"]:
            assert self.actor_num == 1, "when running actor and learner the same time, the actor number should be 1, but received {}".format(
                self.actor_num)
        # dir
        self.model_dir = args.model_dir
        if hasattr(args, "save_dir"):
            self.save_dir = args.save_dir

        self.args = args

    def _inner_loop(
            self,
    ) -> None:
        rollout_infos = self.actor_rollout()
        train_infos = self.learner_update()
        self.post_process()

        self.total_num_steps = (self.episode + 1) * \
                          self.episode_length * self.n_rollout_threads

        # rollout_infos can only be used when env is wrapped with VevMonitor
        self.logger.log_info(rollout_infos,step=self.total_num_steps)
        self.logger.log_info(train_infos,step=self.total_num_steps)


    def reset_and_buffer_init(self):
        # reset env
        returns = self.envs.reset()

        if isinstance(returns, tuple):
            assert len(returns) == 2, "length of env reset returns must be 2, but get {}".format(len(returns))
            obs, info = returns
        else:
            obs = returns

        share_obs = obs

        self.buffer.init_buffer(share_obs.copy(), obs.copy())

    def add2buffer(self, data):
        obs, rewards, dones, infos, values, actions, action_log_probs, rnn_states, rnn_states_critic = data

        rnn_states[dones == True] = np.zeros(((dones == True).sum(), self.recurrent_N, self.hidden_size),
                                             dtype=np.float32)

        rnn_states_critic[dones == True] = np.zeros(
            ((dones == True).sum(), *self.buffer.data.rnn_states_critic.shape[3:]), dtype=np.float32)
        masks = np.ones((self.n_rollout_threads, self.num_agents, 1), dtype=np.float32)
        masks[dones == True] = np.zeros(((dones == True).sum(), 1), dtype=np.float32)

        share_obs = obs

        self.buffer.insert(share_obs, obs, rnn_states, rnn_states_critic, actions, action_log_probs, values, rewards,
                           masks)

    def actor_rollout(self):

        self.trainer.prep_rollout()

        for step in range(self.episode_length):
            # Sample actions
            values, actions, action_log_probs, rnn_states, rnn_states_critic, actions_env = self.act(step)
            # Obser reward and next obs

            actions_env = np.argmax(actions_env[0][0])

            obs, rewards, dones, infos = self.envs.step(actions_env)

            data = obs, rewards, dones, infos, values, actions, action_log_probs, rnn_states, rnn_states_critic

            self.add2buffer(data)

        if self.envs.use_monitor:
            return self.envs.statistics()
        else:
            return {}


    def run(
            self
    ) -> None:
        episodes = int(self.num_env_steps) // self.episode_length // self.learner_n_rollout_threads
        self.episodes = episodes

        self.reset_and_buffer_init()

        for episode in range(episodes):
            self.logger.info("Episode: {}/{}".format(episode, episodes))
            self.episode = episode
            self._inner_loop()

    def learner_update(self):
        if self.use_linear_lr_decay:
            self.trainer.algo_module.lr_decay(self.episode, self.episodes)

        self.compute_returns()
        train_infos = self.train()
        return train_infos

    @torch.no_grad()
    def compute_returns(self):
        self.trainer.prep_rollout()

        next_values = self.trainer.algo_module.get_values(self.buffer.data.get_batch_data("share_obs", -1),
                                                          np.concatenate(
                                                              self.buffer.data.rnn_states_critic[-1]),
                                                          np.concatenate(self.buffer.data.masks[-1]))

        next_values = np.array(
            np.split(
                _t2n(next_values),
                self.learner_n_rollout_threads))
        if "critic" in self.trainer.algo_module.models and isinstance(
                self.trainer.algo_module.models["critic"], DistributedDataParallel):
            value_normalizer = self.trainer.algo_module.models["critic"].module.value_normalizer
        elif "model" in self.trainer.algo_module.models and isinstance(self.trainer.algo_module.models["model"],
                                                                       DistributedDataParallel):
            value_normalizer = self.trainer.algo_module.models["model"].module.value_normalizer
        else:
            value_normalizer = self.trainer.algo_module.get_critic_value_normalizer()
        self.buffer.compute_returns(next_values, value_normalizer)

    def train(self):
        self.trainer.prep_training()
        train_infos = self.trainer.train(self.buffer.data)
        return train_infos

    def post_process(self):
        self.buffer.after_update()

    @torch.no_grad()
    def act(
            self,
            step: int,
    ):
        self.trainer.prep_rollout()

        value, action, action_log_prob, rnn_states, rnn_states_critic \
            = self.trainer.algo_module.get_actions(self.buffer.data.get_batch_data("share_obs", step),
                                                   self.buffer.data.get_batch_data("obs", step),
                                                   np.concatenate(self.buffer.data.rnn_states[step]),
                                                   np.concatenate(self.buffer.data.rnn_states_critic[step]),
                                                   np.concatenate(self.buffer.data.masks[step]))

        # [self.envs, agents, dim]
        values = np.array(np.split(_t2n(value), self.n_rollout_threads))
        actions = np.array(np.split(_t2n(action), self.n_rollout_threads))
        action_log_probs = np.array(np.split(_t2n(action_log_prob), self.n_rollout_threads))
        rnn_states = np.array(np.split(_t2n(rnn_states), self.n_rollout_threads))
        rnn_states_critic = np.array(np.split(_t2n(rnn_states_critic), self.n_rollout_threads))
        # rearrange action
        if self.envs.action_space.__class__.__name__ == 'MultiDiscrete':
            for i in range(self.envs.action_space.shape):
                uc_actions_env = np.eye(self.envs.action_space.high[i] + 1)[actions[:, :, i]]
                if i == 0:
                    actions_env = uc_actions_env
                else:
                    actions_env = np.concatenate((actions_env, uc_actions_env), axis=2)
        elif self.envs.action_space.__class__.__name__ == 'Discrete':
            actions_env = np.squeeze(np.eye(self.envs.action_space.n)[actions], 2)
        else:
            actions_env = actions

        return values, actions, action_log_probs, rnn_states, rnn_states_critic, actions_env



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2023 The TARTRL Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""""""
from abc import abstractmethod
from typing import Dict, Union

import torch
from gym import spaces
from pathlib import Path

from tartrl.modules.base_module import BaseModule
from tartrl.modules.model_config import ModelTrainConfig


class RLModule(BaseModule):
    def __init__(self,
                 args,
                 model_configs: Dict[str, ModelTrainConfig],
                 act_space: spaces.Box,
                 rank: int = 0,
                 world_size: int = 1,
                 device: Union[str, torch.device] = "cpu",
                 ) -> None:

        super(RLModule, self).__init__(args)

        if isinstance(device, str):
            device = torch.device(device)

        self.device = device
        self.lr = args.lr
        self.critic_lr = args.critic_lr
        self.opti_eps = args.opti_eps
        self.weight_decay = args.weight_decay
        self.load_optimizer = args.load_optimizer

        self.act_space = act_space

        self.program_type = args.program_type
        self.rank = rank
        self.world_size = world_size
        self.use_new_network = hasattr(args, "use_new_network") and args.use_new_network

        use_half_actor = self.program_type == "actor" and args.use_half_actor

        for model_key in model_configs:
            model_cg = model_configs[model_key]
            model = model_cg["model"](
                args=args,
                input_space=model_cg["input_space"],
                action_space=act_space,
                device=device,
                use_half=use_half_actor,
            )
            self.models.update({model_key: model})

            if self.program_type == "actor":
                continue

            optimizer = torch.optim.Adam(
                model.parameters(),
                lr=model_cg["lr"],
                eps=args.opti_eps,
                weight_decay=args.weight_decay,
            )
            self.optimizers.update({model_key: optimizer})

            if args.use_amp:
                self.scaler = torch.cuda.amp.GradScaler()
            else:
                self.scaler = None

    @abstractmethod
    def get_actions(self, share_obs, obs, rnn_states_actor, rnn_states_critic, masks, available_actions=None,
                    deterministic=False):

        raise NotImplementedError

    @abstractmethod
    def get_values(self, share_obs, rnn_states_critic, masks):
        raise NotImplementedError

    @abstractmethod
    def evaluate_actions(self, share_obs, obs, rnn_states_actor, rnn_states_critic, action, masks,
                         available_actions=None, active_masks=None, critic_masks_batch=None):
        raise NotImplementedError

    @abstractmethod
    def act(self, obs, rnn_states_actor, masks, available_actions=None, deterministic=False):
        raise NotImplementedError

    @abstractmethod
    def get_critic_value_normalizer(self):
        raise NotImplementedError

    def restore(self, model_dir: str) -> None:
        model_dir = Path(model_dir)
        assert model_dir.exists(), "can not find model directory to restore: {}".format(model_dir)

        for model_name in self.models:
            state_dict = torch.load(str(model_dir) + '/{}.pt'.format(model_name), map_location=self.device)
            self.models[model_name].load_state_dict(state_dict)
            del state_dict

        if self.load_optimizer:
            if Path(str(model_dir) + '/actor_optimizer.pt').exists():
                for optimizer_name in self.optimizers:
                    state_dict = torch.load(str(model_dir) + '/{}_optimizer.pt'.format(optimizer_name),
                                            map_location=self.device)
                    self.optimizers[optimizer_name].load_state_dict(state_dict)
                    del state_dict
            else:
                print("can't find optimizer to restore")
        # TODO
        # optimizer.load_state_dict(resume_state['optimizer'])

    def save(self, save_dir: str) -> None:
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2023 The TARTRL Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""""""

import os
from setuptools import setup, find_packages
import setuptools

def get_install_requires()->list:
    return [
        "setuptools>=50.0",
        "gymnasium",
        "click",
        "termcolor",
        "gym",
        "torch",
        "treevalue",
        "rich",
        "wandb"
    ]

def get_version() -> str:
    # https://packaging.python.org/guides/single-sourcing-package-version/
    init = open(os.path.join("tartrl", "__init__.py"), "r").read().split()
    return init[init.index("__VERSION__") + 2][1:-1]


setup(
    name="tartrl",
    version=get_version(),
    description="reinforcement learning framework",
    long_description=open("README.md", encoding="utf8").read(),
    long_description_content_type="text/markdown",
    author="tartrl contributors",
    author_email="contact@tartrl.cn",
    url="https://github.com/TARTRL/TARTRL",
    packages=setuptools.find_packages(),
    entry_points={'console_scripts': ['tartrl=tartrl.cli.cli:run']},
    project_urls={
        "Code": "https://github.com/TARTRL/TARTRL",
        "Documentation": "https://tartrl-docs.readthedocs.io/zh/latest/",
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    keywords="reinforcement-learning multi-agent "
    "reinforcement-learning-algorithms pytorch machine-learning "
             "baselines toolbox python data-science gym gymnasium",
    python_requires='>=3.8',
    install_requires=get_install_requires(),
)



from .replay_data import ReplayData


class NormalReplayBuffer(object):
    def __init__(self, args, num_agents, obs_space, share_obs_space, act_space, data_client, episode_length=None):
        if episode_length is None:
            episode_length = args.episode_length
        self.data = ReplayData(args, num_agents, obs_space, share_obs_space, act_space, data_client, episode_length)

    def init_buffer(self, share_obs, obs,available_actions=None):
        self.data.init_buffer(share_obs, obs,available_actions)

    def dict_insert(self,data):
        self.data.dict_insert(data)

    def insert(self, share_obs, obs, rnn_states, rnn_states_critic, actions, action_log_probs,
               value_preds, rewards, masks, bad_masks=None, active_masks=None, available_actions=None):
        self.data.insert(share_obs, obs, rnn_states, rnn_states_critic, actions, action_log_probs,
                                 value_preds, rewards, masks, bad_masks, active_masks, available_actions)

    def chooseinsert(self, share_obs, obs, rnn_states, rnn_states_critic, actions, action_log_probs,
                     value_preds, rewards, masks, bad_masks=None, active_masks=None, available_actions=None):
        self.data.chooseinsert(share_obs, obs, rnn_states, rnn_states_critic, actions, action_log_probs,
                                       value_preds, rewards, masks, bad_masks, active_masks, available_actions)

    def after_update(self):
        self.data.after_update()

    def chooseafter_update(self):
        self.data.chooseafter_update()

    def compute_returns(self, next_value, value_normalizer=None):
        self.data.compute_returns(next_value, value_normalizer)

    def feed_forward_generator(self, advantages, num_mini_batch=None, mini_batch_size=None,share_obs_process_func=None):
        return self.data.feed_forward_generator(advantages, num_mini_batch, mini_batch_size,share_obs_process_func=share_obs_process_func)

    def feed_forward_share_obs_generator(self, advantages, num_mini_batch=None, mini_batch_size=None,share_obs_process_func=None):
        # 只包含share_obs和action
        return self.data.feed_forward_share_obs_generator(advantages, num_mini_batch, mini_batch_size,share_obs_process_func=share_obs_process_func)

    def naive_recurrent_generator(self, advantages, num_mini_batch):
        return self.data.naive_recurrent_generator(advantages, num_mini_batch)

    def recurrent_generator(self, advantages, num_mini_batch, data_chunk_length):
        return self.data.recurrent_generator(advantages, num_mini_batch, data_chunk_length)


import math
from typing import Dict, Union, Tuple
import dataclasses
import itertools
from collections.abc import Mapping

import numpy as np
import torch
import torch.distributed as dist
import torch.nn as nn
import torch.nn.functional as F
import time

@dataclasses.dataclass
class ActionIndex:
    start: int
    end: int


def get_action_indices(act_dims):
    curr_action_index = 0
    action_indices = []
    for dim in act_dims:
        action_indices.append(ActionIndex(curr_action_index, curr_action_index + dim))
        curr_action_index += dim
    return action_indices

def mlp(sizes, activation=nn.ReLU, layernorm=True):
    layers = []
    for j in range(len(sizes) - 1):
        layers += [nn.Linear(sizes[j], sizes[j + 1]), activation()]
        if layernorm:
            layers += [nn.LayerNorm([sizes[j + 1]])]
    return nn.Sequential(*layers)

def make_models_for_obs(obs_dim: Dict[str, Union[int, Tuple]], hidden_dim: int, activation,
                        cnn_layers: Dict[str, Tuple]):
    obs_embd_dict = nn.ModuleDict()
    for k, v in obs_dim.items():
        if isinstance(v, int):
            obs_embd_dict.update({
                k: nn.Sequential(nn.LayerNorm([v]),
                                 mlp([v, hidden_dim], activation=activation, layernorm=True))
            })
        else:
            raise NotImplementedError()
    return obs_embd_dict


class RunningMeanStd(nn.Module):

    def __init__(self, input_shape, beta=0.999, epsilon=1e-5):
        super().__init__()
        self.__beta = beta
        self.__eps = epsilon
        self.__input_shape = input_shape

        self.__mean = nn.Parameter(torch.zeros(input_shape), requires_grad=False)
        self.__mean_sq = nn.Parameter(torch.zeros(input_shape), requires_grad=False)
        self.__debiasing_term = nn.Parameter(torch.zeros(1), requires_grad=False)

        self.reset_parameters()

    def reset_parameters(self):
        self.__mean.zero_()
        self.__mean_sq.zero_()
        self.__debiasing_term.zero_()

    def forward(self, *args, **kwargs):
        # we don't implement the forward function because its meaning
        # is somewhat ambiguous
        raise NotImplementedError

    def __check(self, x):
        assert isinstance(x, torch.Tensor)
        trailing_shape = x.shape[-len(self.__input_shape):]
        assert trailing_shape == self.__input_shape, (
            'Trailing shape of input tensor'
            f'{x.shape} does not equal to configured input shape {self.__input_shape}')

    @torch.no_grad()
    def update(self, x):
        self.__check(x)
        norm_dims = tuple(range(len(x.shape) - len(self.__input_shape)))

        batch_mean = x.mean(dim=norm_dims)
        batch_sq_mean = x.square().mean(dim=norm_dims)
        if dist.is_initialized():
            world_size = dist.get_world_size()
            dist.all_reduce(batch_mean)
            dist.all_reduce(batch_sq_mean)
            batch_mean /= world_size
            batch_sq_mean /= world_size

        self.__mean.mul_(self.__beta).add_(batch_mean * (1.0 - self.__beta))
        self.__mean_sq.mul_(self.__beta).add_(batch_sq_mean * (1.0 - self.__beta))
        self.__debiasing_term.mul_(self.__beta).add_(1.0 * (1.0 - self.__beta))

    @torch.no_grad()
    def mean_std(self):
        debiased_mean = self.__mean / self.__debiasing_term.clamp(min=self.__eps)
        debiased_mean_sq = self.__mean_sq / self.__debiasing_term.clamp(min=self.__eps)
        debiased_var = (debiased_mean_sq - debiased_mean**2).clamp(min=1e-2)
        return debiased_mean, debiased_var.sqrt()

    @torch.no_grad()
    def normalize(self, x):
        self.__check(x)
        mean, std = self.mean_std()
        return (x - mean) / std

    @torch.no_grad()
    def denormalize(self, x):
        self.__check(x)
        mean, std = self.mean_std()
        return x * std + mean


def mlp(sizes, activation=nn.ReLU, layernorm=True):
    layers = []
    for j in range(len(sizes) - 1):
        layers += [nn.Linear(sizes[j], sizes[j + 1]), activation()]
        if layernorm:
            layers += [nn.LayerNorm([sizes[j + 1]])]
    return nn.Sequential(*layers)

class AutoResetRNN(nn.Module):

    def __init__(self, input_dim, output_dim, num_layers=1, batch_first=False, rnn_type='gru'):
        super().__init__()
        self.__type = rnn_type
        # print(batch_first)
        # import pdb;pdb.set_trace()
        if self.__type == 'gru':
            self.__net = nn.GRU(input_dim, output_dim, num_layers=num_layers, batch_first=batch_first)
        elif self.__type == 'lstm':
            self.__net = nn.LSTM(input_dim, output_dim, num_layers=num_layers, batch_first=batch_first)
        else:
            raise NotImplementedError(f'RNN type {self.__type} has not been implemented.')
        self._recurrent_N = num_layers

    def __forward(self, x, h):
        if self.__type == 'lstm':
            h = torch.split(h, h.shape[-1] // 2, dim=-1)
            h = (h[0].contiguous(), h[1].contiguous())
        x_, h_ = self.__net(x, h)
        if self.__type == 'lstm':
            h_ = torch.cat(h_, -1)
        return x_, h_

    def forward(self, x, hxs, masks=None):

        if x.size(0) == hxs.size(0):
            x, hxs = self.__forward(x.unsqueeze(0), (hxs * masks.repeat(1, self._recurrent_N).unsqueeze(-1)).transpose(0, 1).contiguous())
            #x= self.gru(x.unsqueeze(0))
            x = x.squeeze(0)
            # hxs = hxs.transpose(0, 1)
        else:
            # x is a (T, N, -1) tensor that has been flatten to (T * N, -1)
            N = hxs.size(0)
            T = int(x.size(0) / N)

            # unflatten
            x = x.view(T, N, x.size(1))

            # Same deal with masks
            masks = masks.view(T, N)

            # Let's figure out which steps in the sequence have a zero for any agent
            # We will always assume t=0 has a zero in it as that makes the logic cleaner
            has_zeros = ((masks[1:] == 0.0)
                         .any(dim=-1)
                         .nonzero()
                         .squeeze()
                         .cpu())

            # +1 to correct the masks[1:]
            if has_zeros.dim() == 0:
                # Deal with scalar
                has_zeros = [has_zeros.item() + 1]
            else:
                has_zeros = (has_zeros + 1).numpy().tolist()

            # add t=0 and t=T to the list
            has_zeros = [0] + has_zeros + [T]

            hxs = hxs.transpose(0, 1)

            outputs = []
            for i in range(len(has_zeros) - 1):
                # We can now process steps that don't have any zeros in masks together!
                # This is much faster
                start_idx = has_zeros[i]
                end_idx = has_zeros[i + 1]
                temp = (hxs * masks[start_idx].view(1, -1, 1).repeat(self._recurrent_N, 1, 1)).contiguous()
                rnn_scores, hxs = self.__forward(x[start_idx:end_idx], temp)
                outputs.append(rnn_scores)

            # assert len(outputs) == T
            # x is a (T, N, -1) tensor
            x = torch.cat(outputs, dim=0)

            # flatten
            x = x.reshape(T * N, -1)
            # hxs = hxs.transpose(0, 1)
        return x, hxs
        # if on_reset is None:
        #     x_, h_ = self.__forward(x, h)
        # else:
        #     outputs = []
        #     for t in range(on_reset.shape[0]):
        #         x_, h = self.__forward(x[t:t + 1], (h * (1 - on_reset[t:t + 1])).contiguous())
        #         outputs.append(x_)
        #     x_ = torch.cat(outputs, 0)
        #     h_ = h
        # return x_, h_


class RecurrentBackbone(nn.Module):

    @property
    def feature_dim(self):
        return self.__feature_dim

    def __init__(
            self,
            obs_dim: int,
            dense_layers: int,
            hidden_dim: int,
            rnn_type: str,
            num_rnn_layers: int,
            dense_layer_gain: float = math.sqrt(2),
            activation='relu',
            layernorm=True,
            batch_first=True
    ):
        super(RecurrentBackbone, self).__init__()

        if activation == 'relu':
            act_fn = nn.ReLU
        elif activation == 'tanh':
            act_fn = nn.Tanh
        elif activation == 'elu':
            act_fn = nn.ELU
        elif activation == 'gelu':
            act_fn = nn.GELU
        else:
            raise NotImplementedError(f"Activation function {activation} not implemented.")

        self.__feature_dim = hidden_dim
        self.__rnn_type = rnn_type
        self.fc = mlp([obs_dim, *([hidden_dim] * dense_layers)], act_fn, layernorm=layernorm)
        for k, p in self.fc.named_parameters():
            if 'weight' in k and len(p.data.shape) >= 2:
                # filter out layer norm weights
                nn.init.orthogonal_(p.data, gain=dense_layer_gain)
            if 'bias' in k:
                nn.init.zeros_(p.data)

        self.num_rnn_layers = num_rnn_layers
        if self.num_rnn_layers:
            self.rnn = AutoResetRNN(hidden_dim,
                                            hidden_dim,
                                            num_layers=num_rnn_layers,
                                            rnn_type=self.__rnn_type,batch_first=batch_first)
            self.rnn_norm = nn.LayerNorm([hidden_dim])
            for k, p in self.rnn.named_parameters():
                if 'weight' in k and len(p.data.shape) >= 2:
                    # filter out layer norm weights
                    nn.init.orthogonal_(p.data)
                if 'bias' in k:
                    nn.init.zeros_(p.data)

    def forward(self, obs, hx, on_reset=None):
        features = self.fc(obs)
        if self.num_rnn_layers > 0:
            features, hx = self.rnn(features, hx, on_reset)
            features = self.rnn_norm(features)
        return features, hx




def _is_dataclass_instance(obj):
    return dataclasses.is_dataclass(obj) and not isinstance(obj, type)


def is_namedarray_instance(obj):
    flag = _is_dataclass_instance(obj) and hasattr(obj, '_fields')
    flag = flag and hasattr(obj, 'items') and issubclass(type(obj), Mapping)
    return flag and not isinstance(obj, type)


def __namedarray_op(op):

    def fn(self, value):
        if not (_is_dataclass_instance(value) and  # Check for matching structure.
                getattr(value, "_fields", None) == self._fields):
            if not _is_dataclass_instance(value):
                value = tuple(None if s is None else value for s in self)
            else:
                raise ValueError('namedarray - set an item with a different data structure')
        try:
            xs = []
            for j, (s, v) in enumerate(zip(self, value)):
                if s is not None and v is not None:
                    exec(f"xs.append(s {op} v)")
        except (ValueError, IndexError, TypeError) as e:
            raise Exception(f"{type(e).__name__} occured in {self.__class__.__name__}"
                            " at field "
                            f"'{self._fields[j]}': {e}") from e
        return type(self)(*xs)

    return fn


def __namedarray_iop(iop):

    def fn(self, value):
        if not (_is_dataclass_instance(value) and  # Check for matching structure.
                getattr(value, "_fields", None) == self._fields):
            if not _is_dataclass_instance(value):
                value = {k: None if s is None else value for k, s in self.items()}
            else:
                raise ValueError('namedarray - set an item with a different data structure')
        try:
            for j, (k, v) in enumerate(zip(self.keys(), value.values())):
                if self[k] is not None and v is not None:
                    exec(f"self[k] {iop} v")
        except (ValueError, IndexError, TypeError) as e:
            raise Exception(f"{type(e).__name__} occured in {self.__class__.__name__}"
                            " at field "
                            f"'{self._fields[j]}': {e}") from e
        return self

    return fn


def namedarray(cls, *args, **kwargs):
    data_cls = dataclasses.dataclass(cls, *args, **kwargs)
    typename = data_cls.__class__.__name__

    def __iter__(self):

        def gen():
            for k in self.__dataclass_fields__.keys():
                yield getattr(self, k)

        return gen()

    def __getitem__(self, loc):
        if isinstance(loc, str):
            # str indexing like in dict
            return getattr(self, loc)
        else:
            try:
                return type(self)(*(None if s is None else s[loc] for s in self))
            except IndexError as e:
                for j, s in enumerate(self):
                    if s is None:
                        continue
                    try:
                        _ = s[loc]
                    except IndexError:
                        raise Exception(f"Occured in {self.__class__} at field "
                                        f"'{self._fields[j]}'.") from e

    def __setitem__(self, loc, value):
        if isinstance(loc, str):
            setattr(self, loc, value)
        else:
            if not (_is_dataclass_instance(value) and  # Check for matching structure.
                    getattr(value, "_fields", None) == self._fields):
                if not _is_dataclass_instance(value):
                    # Repeat value for each but respect any None.
                    value = tuple(None if s is None else value for s in self)
                else:
                    raise ValueError('namedarray - set an item with a different data structure')
            try:
                for j, (s, v) in enumerate(zip(self, value)):
                    if s is not None and v is not None:
                        s[loc] = v
            except (ValueError, IndexError, TypeError) as e:
                raise Exception(f"{type(e).__name__} occured in {self.__class__.__name__}"
                                " at field "
                                f"'{self._fields[j]}': {e}") from e

    def __contains__(self, key):
        return key in self._fields

    def values(self):
        for v in self:
            yield v

    def keys(self):
        for k in self._fields:
            yield k

    def items(self):
        for k, v in zip(self._fields, self):
            yield k, v

    def to_dict(self):
        result = {}
        for k, v in self.items():
            if is_namedarray_instance(v):
                result[k] = v.to_dict()
            elif v is None:
                result[k] = None
            else:
                result[k] = v
        return result

    @property
    def shape(self):
        return recursive_apply(self, lambda x: x.shape).to_dict()

    def size(self):
        return self.shape

    methods = [__getitem__, __setitem__, __iter__, __contains__, values, keys, items]

    for method in methods:
        method.__qualname__ = f'{typename}.{method.__name__}'

    ops = {
        '__add__': __namedarray_op('+'),
        '__sub__': __namedarray_op('-'),
        '__mul__': __namedarray_op('*'),
        '__truediv__': __namedarray_op('/')
    }
    iops = {
        '__iadd__': __namedarray_iop('+='),
        '__isub__': __namedarray_iop('-='),
        '__imul__': __namedarray_iop('*='),
        '__itruediv__': __namedarray_iop('/=')
    }
    for name, op in itertools.chain(ops.items(), iops.items()):
        op.__qualname__ = f'{typename}.{name}'

    arg_list = repr(list(data_cls.__dataclass_fields__.keys())).replace("'", "")[1:-1]
    class_namespace = {
        '__doc__': f'{typename}({arg_list})',
        '__slots__': (),
        '__iter__': __iter__,
        '_fields': list(data_cls.__dataclass_fields__.keys()),
        '__getitem__': __getitem__,
        '__setitem__': __setitem__,
        '__contains__': __contains__,
        'items': items,
        'keys': keys,
        'values': values,
        'to_dict': to_dict,
        'shape': shape,
        'size': size,
    }
    class_namespace = {**class_namespace, **ops, **iops}
    for k, v in class_namespace.items():
        setattr(data_cls, k, v)

    Mapping.register(data_cls)
    return data_cls


def array_like(x, value=0):
    if is_namedarray_instance(x):
        return type(x)(*[array_like(xx, value) for xx in x])
    else:
        if isinstance(x, np.ndarray):
            data = np.zeros_like(x)
        else:
            assert isinstance(x, torch.Tensor), ('Currently, namedarray only supports'
                                                 f' torch.Tensor and numpy.array (input is {type(x)})')
            data = torch.zeros_like(x)
        if value != 0:
            data[:] = value
        return data


def __array_filter_none(xs):
    is_not_nones = [x is not None for x in xs]
    if all(is_not_nones) or all(x is None for x in xs):
        return
    else:
        example_x = xs[is_not_nones.index(True)]
        for i, x in enumerate(xs):
            xs[i] = array_like(example_x) if x is None else x


def recursive_aggregate(xs, aggregate_fn):
    __array_filter_none(xs)
    assert all([type(x) == type(xs[0]) for x in xs]), ([type(x) for x in xs], xs)

    if is_namedarray_instance(xs[0]):
        return type(xs[0])(*[recursive_aggregate([x[k] for x in xs], aggregate_fn) for k in xs[0].keys()])
    elif xs[0] is None:
        return None
    else:
        return aggregate_fn(xs)


def recursive_apply(x, fn):
    if is_namedarray_instance(x):
        return type(x)(*[recursive_apply(v, fn) for v in x.values()])
    elif x is None:
        return None
    else:
        return fn(x)

class Action:
    pass

@namedarray
class DiscreteAction(Action):
    x: np.ndarray

    def __eq__(self, other):
        assert isinstance(other, DiscreteAction), \
            "Cannot compare DiscreteAction to object of class{}".format(other.__class__.__name__)
        return self.key == other.key

    def __hash__(self):
        return hash(self.x.item())

    @property
    def key(self):
        return self.x.item()

class PopArtValueHead(nn.Module):

    def __init__(self, input_dim, critic_dim, beta=0.999, epsilon=1e-5, burn_in_updates=torch.inf):
        super().__init__()
        self.__rms = RunningMeanStd((critic_dim,), beta, epsilon)

        self.__weight = nn.Parameter(torch.zeros(critic_dim, input_dim))
        nn.init.orthogonal_(self.__weight)
        self.__bias = nn.Parameter(torch.zeros(critic_dim))

        self.__burn_in_updates = burn_in_updates
        self.__update_cnt = 0

    def forward(self, feature):
        return F.linear(feature, self.__weight, self.__bias)

    @torch.no_grad()
    def update(self, x):
        old_mean, old_std = self.__rms.mean_std()
        self.__rms.update(x)
        new_mean, new_std = self.__rms.mean_std()
        self.__update_cnt += 1

        if self.__update_cnt > self.__burn_in_updates:
            self.__weight.data[:] = self.__weight * (old_std / new_std).unsqueeze(-1)
            self.__bias.data[:] = (old_std * self.__bias + old_mean - new_mean) / new_std

    @torch.no_grad()
    def normalize(self, x):
        return self.__rms.normalize(x)

    @torch.no_grad()
    def denormalize(self, x):
        return self.__rms.denormalize(x)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2023 The TARTRL Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""""""

from typing import Union, Tuple, Dict, Optional

import numpy as np
import gym
import torch

from tartrl.modules.ppo_module import PPOModule
from tartrl.configs.config import build_parser
from tartrl.modules.common.base_net import BaseNet

class PPONet(BaseNet):
    def __init__(
            self,
            env: Union[gym.Env, str],
            device: Union[torch.device, str] = 'cpu',
            n_rollout_threads: int = 1,
    ) -> None:
        super().__init__()
        parser = build_parser()
        args = parser.parse_known_args("")[0]
        args.n_rollout_threads = n_rollout_threads
        args.learner_n_rollout_threads = args.n_rollout_threads

        if args.rnn_type == "gru":
            rnn_hidden_size = args.hidden_size
        elif args.rnn_type == "lstm":
            rnn_hidden_size =args.hidden_size * 2
        else:
            raise NotImplementedError(
                f'RNN type {args.rnn_type} has not been implemented.')
        args.rnn_hidden_size = rnn_hidden_size

        if isinstance(device, str):
            device = torch.device(device)

        self.module = PPOModule(
            args=args,
            policy_input_space=env.observation_space,
            critic_input_space=env.observation_space,
            act_space=env.action_space,
            share_model=False,
            device=device,
            rank=0,
            world_size=1,
        )

        self.args = args
        self.env = env
        self.device = device
        self.rnn_states_actor = None
        self.masks = None

    def act(
            self,
            observation: Union[np.ndarray, Dict[str, np.ndarray]],
            deterministic: bool = False,
    ) -> Tuple[np.ndarray, Optional[Tuple[np.ndarray, ...]]]:

        if not self.first_reset:
            self.reset()

        actions, self.rnn_states_actor = self.module.act(
            obs = observation,
            rnn_states_actor=self.rnn_states_actor,
            masks=self.masks,
            available_actions=None,
            deterministic=deterministic,
        )

        actions = actions.item()

        return actions, self.rnn_states_actor

    def reset(self):
        self.first_reset = False
        self.rnn_states_actor, self.masks = self.module.init_rnn_states(
            rollout_num=self.args.n_rollout_threads,
            agent_num=self.env.agent_num,
            rnn_layers=self.args.recurrent_N,
            hidden_size=self.args.rnn_hidden_size,
        )




from typing import Union

import numpy as np
import torch
import torch.nn as nn
from torch.nn.parallel import DistributedDataParallel

from tartrl.algorithms.base_algorithm import BaseAlgorithm
from tartrl.utils.util import check
from tartrl.modules.utils.util import get_gard_norm, huber_loss, mse_loss

from tartrl.modules.networks.utils.distributed_utils import reduce_tensor


class PPOAlgorithm(BaseAlgorithm):
    def __init__(
            self,
            args,
            init_module,
            device: Union[str, torch.device] = "cpu",
    ) -> None:
        self.use_joint_action_loss = args.use_joint_action_loss
        super(PPOAlgorithm, self).__init__(args, init_module, device)

    def ppo_update(self, sample, turn_on=True):
        for optimizer in self.algo_module.optimizers.values():
            optimizer.zero_grad()

        share_obs_batch, obs_batch, rnn_states_batch, rnn_states_critic_batch, actions_batch, \
            value_preds_batch, return_batch, masks_batch, active_masks_batch, old_action_log_probs_batch, \
            adv_targ, available_actions_batch = sample

        old_action_log_probs_batch = check(old_action_log_probs_batch).to(**self.tpdv)
        adv_targ = check(adv_targ).to(**self.tpdv)
        value_preds_batch = check(value_preds_batch).to(**self.tpdv)
        return_batch = check(return_batch).to(**self.tpdv)
        active_masks_batch = check(active_masks_batch).to(**self.tpdv)

        if self.use_amp:
            with torch.cuda.amp.autocast():
                loss_list, value_loss, policy_loss, dist_entropy, ratio = self.prepare_loss(share_obs_batch,
                                                                                            obs_batch,
                                                                                            rnn_states_batch,
                                                                                            rnn_states_critic_batch,
                                                                                            actions_batch,
                                                                                            masks_batch,
                                                                                            available_actions_batch,
                                                                                            old_action_log_probs_batch,
                                                                                            adv_targ, value_preds_batch,
                                                                                            return_batch,
                                                                                            active_masks_batch, turn_on)
            for loss in loss_list:
                self.algo_module.scaler.scale(loss).backward()
        else:
            loss_list, value_loss, policy_loss, dist_entropy, ratio = self.prepare_loss(share_obs_batch,
                                                                                        obs_batch,
                                                                                        rnn_states_batch,
                                                                                        rnn_states_critic_batch,
                                                                                        actions_batch,
                                                                                        masks_batch,
                                                                                        available_actions_batch,
                                                                                        old_action_log_probs_batch,
                                                                                        adv_targ, value_preds_batch,
                                                                                        return_batch,
                                                                                        active_masks_batch, turn_on)
            for loss in loss_list:
                loss.backward()

        if "transformer" in self.algo_module.models:
            if self._use_max_grad_norm:
                grad_norm = nn.utils.clip_grad_norm_(self.algo_module.models['transformer'].parameters(),
                                                     self.max_grad_norm)
            else:
                grad_norm = get_gard_norm(self.algo_module.models['transformer'].parameters())
            critic_grad_norm = grad_norm
            actor_grad_norm = grad_norm

        else:
            if self._use_max_grad_norm:
                actor_grad_norm = nn.utils.clip_grad_norm_(self.algo_module.models['policy'].parameters(),
                                                           self.max_grad_norm)
            else:
                actor_grad_norm = get_gard_norm(self.algo_module.models['policy'].parameters())

            if self._use_max_grad_norm:
                critic_grad_norm = nn.utils.clip_grad_norm_(self.algo_module.models['critic'].parameters(),
                                                            self.max_grad_norm)
            else:
                critic_grad_norm = get_gard_norm(self.algo_module.models['critic'].parameters())

        if self.use_amp:
            for optimizer in self.algo_module.optimizers.values():
                self.algo_module.scaler.unscale_(optimizer)

            for optimizer in self.algo_module.optimizers.values():
                self.algo_module.scaler.step(optimizer)

            self.algo_module.scaler.update()
        else:
            for optimizer in self.algo_module.optimizers.values():
                optimizer.step()

        if self.world_size > 1:
            torch.cuda.synchronize()

        return value_loss, critic_grad_norm, policy_loss, dist_entropy, actor_grad_norm, ratio

    def cal_value_loss(
            self,
            value_normalizer,
            values,
            value_preds_batch,
            return_batch,
            active_masks_batch
    ):
        value_pred_clipped = value_preds_batch + (values - value_preds_batch).clamp(-self.clip_param, self.clip_param)

        if self._use_popart or self._use_valuenorm:
            value_normalizer.update(return_batch)
            error_clipped = value_normalizer.normalize(return_batch) - value_pred_clipped
            error_original = value_normalizer.normalize(return_batch) - values
        else:
            error_clipped = return_batch - value_pred_clipped
            error_original = return_batch - values

        if self._use_huber_loss:
            value_loss_clipped = huber_loss(error_clipped, self.huber_delta)
            value_loss_original = huber_loss(error_original, self.huber_delta)
        else:
            value_loss_clipped = mse_loss(error_clipped)
            value_loss_original = mse_loss(error_original)

        if self._use_clipped_value_loss:
            value_loss = torch.max(value_loss_original, value_loss_clipped)
        else:
            value_loss = value_loss_original

        if self._use_value_active_masks:
            value_loss = (value_loss * active_masks_batch).sum() / active_masks_batch.sum()
        else:
            value_loss = value_loss.mean()

        return value_loss

    def to_single_np(
            self,
            input
    ):
        reshape_input = input.reshape(-1, self._num_agents, *input.shape[1:])
        return reshape_input[:, 0, ...]

    def prepare_loss(
            self,
            share_obs_batch,
            obs_batch,
            rnn_states_batch,
            rnn_states_critic_batch,
            actions_batch,
            masks_batch,
            available_actions_batch,
            old_action_log_probs_batch,
            adv_targ,
            value_preds_batch,
            return_batch,
            active_masks_batch,
            turn_on
    ):
        loss_list = []

        if self.use_joint_action_loss:
            share_obs_batch = self.to_single_np(share_obs_batch)
            rnn_states_critic_batch = self.to_single_np(rnn_states_critic_batch)
            critic_masks_batch = self.to_single_np(masks_batch)
            value_preds_batch = self.to_single_np(value_preds_batch)
            return_batch = self.to_single_np(return_batch)
            adv_targ = adv_targ.reshape(-1, self._num_agents, 1)
            adv_targ = adv_targ[:, 0, :]

        else:
            critic_masks_batch = masks_batch

        values, action_log_probs, dist_entropy, policy_values = \
            self.algo_module.evaluate_actions(
                share_obs_batch,
                obs_batch,
                rnn_states_batch,
                rnn_states_critic_batch,
                actions_batch,
                masks_batch,
                available_actions_batch,
                active_masks_batch,
                critic_masks_batch=critic_masks_batch
            )

        if self.use_joint_action_loss:
            action_log_probs_copy = action_log_probs.reshape(-1, self._num_agents, action_log_probs.shape[-1]).sum(
                dim=(1, -1), keepdim=True).reshape(-1, 1)
            old_action_log_probs_batch_copy = old_action_log_probs_batch.reshape(-1, self._num_agents,
                                                                                 old_action_log_probs_batch.shape[
                                                                                     -1]).sum(dim=(1, -1),
                                                                                              keepdim=True).reshape(-1,
                                                                                                                    1)

            active_masks_batch = active_masks_batch.reshape(-1, self._num_agents, 1)
            active_masks_batch = active_masks_batch[:, 0, :]

            ratio = torch.exp(action_log_probs_copy - old_action_log_probs_batch_copy)
        else:
            ratio = torch.exp(action_log_probs - old_action_log_probs_batch)

        if self.dual_clip_ppo:
            ratio = torch.min(ratio, self.dual_clip_coeff)

        surr1 = ratio * adv_targ
        surr2 = torch.clamp(ratio, 1.0 - self.clip_param, 1.0 + self.clip_param) * adv_targ

        surr_final = torch.min(surr1, surr2)

        if self._use_policy_active_masks:
            policy_action_loss = (-torch.sum(surr_final, dim=-1,
                                             keepdim=True) * active_masks_batch).sum() / active_masks_batch.sum()
        else:
            policy_action_loss = -torch.sum(surr_final, dim=-1, keepdim=True).mean()

        if self._use_policy_vhead:
            if isinstance(self.algo_module.models["actor"], DistributedDataParallel):
                policy_value_normalizer = self.algo_module.models["actor"].module.value_normalizer
            else:
                policy_value_normalizer = self.algo_module.models["actor"].value_normalizer
            policy_value_loss = self.cal_value_loss(policy_value_normalizer, policy_values, value_preds_batch,
                                                    return_batch, active_masks_batch)
            policy_loss = policy_action_loss + policy_value_loss * self.policy_value_loss_coef
        else:
            policy_loss = policy_action_loss

        # critic update
        if isinstance(self.algo_module.models["critic"], DistributedDataParallel):
            value_normalizer = self.algo_module.models["critic"].module.value_normalizer
        else:
            value_normalizer = self.algo_module.get_critic_value_normalizer()
        value_loss = self.cal_value_loss(value_normalizer, values, value_preds_batch, return_batch,
                                         active_masks_batch)

        if "transformer" in self.algo_module.models:
            loss = policy_loss - dist_entropy * self.entropy_coef + value_loss * self.value_loss_coef
            loss_list.append(loss)
        else:
            if turn_on:
                final_p_loss = policy_loss - dist_entropy * self.entropy_coef

                loss_list.append(final_p_loss)
            final_v_loss = value_loss * self.value_loss_coef
            loss_list.append(final_v_loss)
        return loss_list, value_loss, policy_loss, dist_entropy, ratio



    def train(self, buffer, turn_on=True):
        if self._use_popart or self._use_valuenorm:
            if isinstance(self.algo_module.models["critic"], DistributedDataParallel):
                value_normalizer = self.algo_module.models["critic"].module.value_normalizer
            else:
                value_normalizer = self.algo_module.get_critic_value_normalizer()
            advantages = buffer.returns[:-1] - value_normalizer.denormalize(buffer.value_preds[:-1])
        else:
            advantages = buffer.returns[:-1] - buffer.value_preds[:-1]

        if self._use_adv_normalize:
            advantages = (advantages - advantages.mean()) / (
                    advantages.std() + 1e-5)

        advantages_copy = advantages.copy()
        advantages_copy[buffer.active_masks[:-1] == 0.0] = np.nan
        mean_advantages = np.nanmean(advantages_copy)
        std_advantages = np.nanstd(advantages_copy)
        advantages = (advantages - mean_advantages) / (std_advantages + 1e-5)

        train_info = {}

        train_info['value_loss'] = 0
        train_info['policy_loss'] = 0
        if self._use_new_network:
            train_info['id_loss'] = 0
            train_info['id_entropy'] = 0

        train_info['dist_entropy'] = 0
        train_info['actor_grad_norm'] = 0
        train_info['critic_grad_norm'] = 0
        train_info['ratio'] = 0
        if self.world_size > 1:
            train_info['reduced_value_loss'] = 0
            train_info['reduced_policy_loss'] = 0
            if self._use_new_network:
                train_info['reduced_id_loss'] = 0
                train_info['reduced_id_entropy'] = 0

        for _ in range(self.ppo_epoch):
            if 'transformer' in self.algo_module.models:
                data_generator = buffer.feed_forward_generator_transformer(advantages, self.num_mini_batch)
            elif self._use_recurrent_policy:
                if self.use_joint_action_loss:
                    data_generator = buffer.recurrent_generator_v3(advantages, self.num_mini_batch,
                                                                   self.data_chunk_length)
                else:
                    data_generator = buffer.recurrent_generator(advantages, self.num_mini_batch, self.data_chunk_length)
            elif self._use_naive_recurrent:
                data_generator = buffer.naive_recurrent_generator(advantages, self.num_mini_batch)
            else:
                data_generator = buffer.feed_forward_generator(advantages, self.num_mini_batch)

            for sample in data_generator:
                value_loss, critic_grad_norm, policy_loss, dist_entropy, actor_grad_norm, ratio \
                    = self.ppo_update(sample, turn_on)

                if self.world_size > 1:
                    train_info['reduced_value_loss'] += reduce_tensor(value_loss.data, self.world_size)
                    train_info['reduced_policy_loss'] += reduce_tensor(policy_loss.data, self.world_size)

                train_info['value_loss'] += value_loss.item()
                train_info['policy_loss'] += policy_loss.item()

                train_info['dist_entropy'] += dist_entropy.item()
                train_info['actor_grad_norm'] += actor_grad_norm
                train_info['critic_grad_norm'] += critic_grad_norm
                train_info['ratio'] += ratio.mean().item()

        num_updates = self.ppo_epoch * self.num_mini_batch

        for k in train_info.keys():
            train_info[k] /= num_updates

        for optimizer in self.algo_module.optimizers.values():
            if hasattr(optimizer, 'sync_lookahead'):
                optimizer.sync_lookahead()

        return train_info



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2023 The TARTRL Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""""""
import numpy as np

from tartrl.envs import make
from tartrl.modules.common import PPONet as Net
from tartrl.runners.common import PPOAgent as Agent

def train_agent(env: str):
    render_model = "rgb_array"
    env = make(env, render_mode=render_model)
    # 创建 神经网络
    net = Net(env)
    # 初始化训练器
    agent = Agent(net)
    # 开始训练
    agent.train(total_time_steps=20000)

    obs, info = env.reset()
    done = False
    step = 0
    total_reward = 0
    while not done:
        # 智能体根据 observation 预测下一个动作
        action, _ = agent.act(obs, deterministic=True)
        obs, r, done, info = env.step(action)
        total_reward += r
        step += 1
        # print(f"{step}: action:{action}, reward:{np.squeeze(r)}")
    print(f"Total reward: {np.squeeze(total_reward)}")

    env.close()